import { CommandSpec, Command, EntitySpec, DatabaseEntityBuilder, OutputSpec } from "./Command";
import { APISuccessResponse, DatabaseResult, DatabaseResultNotifier, OperationNotifier } from "./Result";
import { Identity } from "./IdentitySource";
export declare type DataSourceHierarchy = {
    [key: string]: EntitySpec;
};
export declare type RequestProcessingFunction = (dsName: string, entity: string, operation: string, args: any[]) => Promise<APISuccessResponse>;
export declare type BatchRequestProcessingFunction = (dsName: string, args: [string, any[]][]) => Promise<APISuccessResponse>;
export declare type ConstructAPIFunction = (hier: DataSourceHierarchy, ns: string, processor: RequestProcessingFunction, batchProcessor: BatchRequestProcessingFunction) => Object;
export interface DataSourceConfig {
    name: string;
    driver: string;
    driverConfig: DriverConfig;
    allowedUsernames?: string[];
    allowedRoles?: string[];
    extensions?: string[];
}
export declare enum HandlerType {
    Database = 0,
    Declared = 1
}
export interface Handler {
    readonly handlerType: HandlerType;
    loadEntitySpecs(): Promise<EntitySpec[]>;
    useAPI(api: any): any;
    executeCommand(command: Command): Promise<APISuccessResponse>;
    executeBatch(commands: Command[]): Promise<APISuccessResponse>;
}
export declare type DriverConfig = {
    [key: string]: any;
};
export declare type DatabaseConnectionRef = any;
export interface Driver {
    readonly handlerType: HandlerType;
}
export interface DatabaseDriver extends Driver {
    openConnection(notifier: OperationNotifier<DatabaseConnectionRef>): void;
    closeConnection(connection: DatabaseConnectionRef): void;
    loadEntitySpecs(connection: DatabaseConnectionRef, builder: DatabaseEntityBuilder): void;
    executeCommand(connection: DatabaseConnectionRef, command: Command, notifier: DatabaseResultNotifier): void;
    executeSQL(connection: DatabaseConnectionRef, sql: string, notifier: DatabaseResultNotifier, outputs: OutputSpec[] | null): void;
    beginTransaction(connection: DatabaseConnectionRef): Promise<void>;
    commitTransaction(connection: DatabaseConnectionRef): Promise<void>;
    rollbackTransaction(connection: DatabaseConnectionRef): Promise<void>;
}
export interface DeclaredDriver extends Driver {
    useNamespaceAPI?(api: any): any;
    readonly API: {
        [key: string]: Function;
    };
}
export declare class DataSourceFinder {
    readonly sources: {
        [key: string]: DataSource;
    };
    constructor();
    addSource(dsconfig: DataSourceConfig, processor: RequestProcessingFunction, batchProcessor: BatchRequestProcessingFunction): Promise<void>;
    hasSourceNamed(name: string): boolean;
    findSourceWithName(name: string): DataSource;
    allowedSources(ident: Identity | null): string[];
}
export declare class DataSource {
    private readonly config;
    private readonly _entities;
    private readonly handlers;
    private readonly extensions;
    protected handler: Handler;
    protected _api: any;
    readonly hierarchy: DataSourceHierarchy;
    readonly api: any;
    constructor(dsconfig: DataSourceConfig);
    constructHandler(driverPath: string, driverConfig: any): Handler;
    getName(): string;
    allowIdentity(ident: Identity): boolean;
    hasEntity(entity: string): boolean;
    hasOperation(entity: string, operation: string): boolean;
    defaultOperationName(entity: string): string;
    specForEntity(entityName: string): EntitySpec;
    handlerForEntity(entityName: string): Handler;
    specForOperation(entityName: string, operation: string): CommandSpec;
    constructAPI(namespace: string, processor: RequestProcessingFunction, batchProcessor: BatchRequestProcessingFunction): Promise<void>;
    executeCommand(entityName: string, operationName: string, args: any[]): Promise<any>;
    executeBatch(args: [string, any[]][]): Promise<APISuccessResponse>;
    executeSQL(sql: string): Promise<DatabaseResult>;
    getAPIJavaScriptConstructor(): string;
}
export declare var constructAPIFromHierarchy: ConstructAPIFunction;
