"use strict";
/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const http = require("http");
const fs = require("fs");
const path = require("path");
const url = require("url");
const DataSource_1 = require("./DataSource");
const Mimes_1 = require("./Mimes");
const IdentitySource_1 = require("./IdentitySource");
const log = require("./Logger");
const TRANSIT_API_VERSION = 1;
var TransitRequestType;
(function (TransitRequestType) {
    TransitRequestType[TransitRequestType["API"] = 1] = "API";
    TransitRequestType[TransitRequestType["Document"] = 2] = "Document";
})(TransitRequestType || (TransitRequestType = {}));
async function startServer(configArg) {
    let config = parseConfig(configArg);
    if (config.eventLog)
        log.configureEvent(config.eventLog);
    if (config.eventLog)
        log.configureAPI(config.apiLog);
    if (config.eventLog)
        log.configureDocument(config.documentLog);
    let runtime = await startRuntime(config);
    let server = http.createServer((req, res) => {
        req.on("error", (err) => {
            log.error(err);
            res.statusCode = 400;
            res.end();
        });
        res.on("error", (err) => log.error(err));
        runtime.processSiteRequest(req, res).catch(err => {
            res.writeHead(500);
            log.error(err);
            res.write(JSON.stringify(err));
            res.end();
        });
        if (req["transitRequestType"] == TransitRequestType.API) {
            log.api(req, res);
        }
        else {
            log.document(req, res);
        }
    }).listen(config.port);
    log.info(`Server started and listening on port ${config.port}`);
    server.runtime = runtime;
    return server;
}
exports.startServer = startServer;
async function startRuntime(configArg) {
    return await new ServerRuntime(parseConfig(configArg)).start();
}
exports.startRuntime = startRuntime;
async function generateAPI(configArg, ns) {
    let runtime = await startRuntime(configArg);
    return await runtime.getAPI(ns);
}
exports.generateAPI = generateAPI;
function parseConfig(configArg) {
    if (typeof configArg == "object") {
        return configArg;
    }
    try {
        return eval(`(${fs.readFileSync(String(configArg)).toString()})`);
    }
    catch (e) {
        throw new Error(`Unable to parse config file ${configArg}: ${e}`);
    }
}
exports.parseConfig = parseConfig;
class ServerRuntime {
    constructor(config) {
        this.identitySource = null;
        this.dataSourceFinder = new DataSource_1.DataSourceFinder();
        this.configFileRegExp = new RegExp(`^/v${TRANSIT_API_VERSION}/\\$clientConfig\.js$`);
        this.namespacesRegExp = new RegExp(`^/v${TRANSIT_API_VERSION}/namespaces$`);
        this.apiFileRegExp = new RegExp(`^/v${TRANSIT_API_VERSION}/(\\w[\\w0-9]*)\\.(json|js|html)$`);
        this.apiPathRegExp = new RegExp(`^/v${TRANSIT_API_VERSION}/(\\w[\\w0-9]*)/([^/]+)(?:/([^/]+))?$`);
        this.apiJSTemplate = '';
        this.apiHTMLTemplate = '';
        this.call = async (dsName, entity, operation, args) => {
            let ds = this.dataSourceFinder.findSourceWithName(dsName);
            if (operation == null) {
                operation = ds.defaultOperationName(entity);
            }
            return await ds.executeCommand(entity, operation, args);
        };
        this.callBatch = async (dsName, data) => {
            return await this.dataSourceFinder.findSourceWithName(dsName).executeBatch(data);
        };
        this.config = config;
        if (this.config.identity != null) {
            this.identitySource = IdentitySource_1.loadIdentitySource(this.config.identity);
        }
        ServerRuntime._instance = this;
    }
    static getInstance() {
        return this._instance;
    }
    async start() {
        if (this.config.namespaces == null || typeof this.config.namespaces != "object") {
            throw new Error("Server configuration must have at least one namespace.");
        }
        if (!fs.existsSync(this.config.contentDir)) {
            log.warn(`Warning: content directory ${this.config.contentDir} not found`);
        }
        for (var config of this.config.namespaces) {
            if (!config.name)
                throw new Error('namespace config must contain a name');
            await this.dataSourceFinder.addSource(config, this.call, this.callBatch);
        }
        this.apiJSTemplate = fs.readFileSync(path.join(path.dirname(module.id), "resource/api_template.js")).toString();
        this.apiHTMLTemplate = fs.readFileSync(path.join(path.dirname(module.id), "resource/api_template.html")).toString();
        return this;
    }
    async processSiteRequest(request, response) {
        let urlInfo = url.parse(request.url, true);
        let pathname = urlInfo.pathname.replace(/\/\//g, "/");
        // make sure to always read the POST body or client-side
        // requests will hang waiting for it to be read.
        let body = "";
        if (request.method == "POST") {
            body = await readPostBody(request);
        }
        // client config "file" request?
        if (pathname.match(this.configFileRegExp)) {
            request['transitRequestType'] = TransitRequestType.Document;
            response.writeHead(200, { "Content-Type": "text/plain" });
            response.end(`transitClientConfig = ${JSON.stringify(this.config.clientConfig)};`);
            return;
        }
        if (pathname.match(this.namespacesRegExp)) {
            request['transitRequestType'] = TransitRequestType.API;
            try {
                let namespaces = await this.authorizedNamespaces(request);
                this.writeSuccessResponse(response, {
                    meta: { handler: DataSource_1.HandlerType.Declared },
                    data: namespaces
                });
            }
            catch (err) {
                response.writeHead(500, { "Content-Type": "application/json" });
                let retErr = this.makeAPIErrorResponse(500, String(err), `Unable to fetch namespaces\n${err.stack || err}`);
                response.end(JSON.stringify(retErr));
            }
            return;
        }
        // API JavaScript "file" request?
        let match = pathname.match(this.apiFileRegExp);
        let [dummy1, jsns, ext] = match || [null, null, null];
        if (jsns && this.dataSourceFinder.hasSourceNamed(jsns)) {
            request['transitRequestType'] = TransitRequestType.Document;
            if (ext == "json") {
                let ds = this.dataSourceFinder.findSourceWithName(jsns);
                response.writeHead(200, { "Content-Type": "application/json" });
                response.end(JSON.stringify({ meta: {}, data: ds.hierarchy }), 'utf-8');
            }
            else if (ext == "js") {
                response.writeHead(200, { "Content-Type": "text/plain" });
                response.end(this.generateAPIJavaScript(jsns), 'utf-8');
            }
            else {
                response.writeHead(200, { "Content-Type": "text/html" });
                response.end(this.generateAPIHTML(jsns), 'utf-8');
            }
            return;
        }
        match = pathname.match(this.apiPathRegExp);
        let [dummy2, ns, entity, operation] = match || [null, null, null, null];
        // valid data source?
        if (ns && this.dataSourceFinder.hasSourceNamed(ns)) {
            request['transitRequestType'] = TransitRequestType.API;
            let ds = this.dataSourceFinder.findSourceWithName(ns);
            let ident = await this.isAuthorized(request, response, ds);
            if (!ident) {
                return;
            }
            try {
                let data;
                try {
                    data = JSON.parse(body);
                }
                catch (e) {
                    throw new Error(`Parse error in JSON: ${e.toString()}\n${body}`);
                }
                if (entity == "$authorize" && operation == undefined) {
                    // we already authorized above!
                    this.writeSuccessResponse(response, {
                        data: ident,
                        meta: { handler: DataSource_1.HandlerType.Declared }
                    });
                    return;
                }
                else if (entity == "$batch-api" && operation == undefined) {
                    // batch request
                    if (!Array.isArray(data)) {
                        response.writeHead(500, { "Content-Type": "application/json" });
                        response.end(JSON.stringify(this.makeAPIErrorResponse(500, `batch POST body must be an array. POST body was:\n${body}`)));
                    }
                    for (let item of data) {
                        let [entity, operation] = String(item[0]).split(".");
                        let err = this.validateAPIRequest(ds, entity, operation, item[1]);
                        if (err) {
                            response.writeHead(err.errors[0].status, { "Content-Type": "application/json" });
                            response.end(JSON.stringify(err));
                            return;
                        }
                    }
                    let result = await this.callBatch(ns, data);
                    this.writeSuccessResponse(response, result);
                    return;
                }
                else {
                    let err = this.validateAPIRequest(ds, entity, operation, data);
                    if (err) {
                        response.writeHead(err.errors[0].status, { "Content-Type": "application/json" });
                        response.end(JSON.stringify(err));
                        return;
                    }
                    if (!data)
                        data = [];
                    let result = await this.call(ns, entity, operation, data);
                    this.writeSuccessResponse(response, result);
                    return;
                }
            }
            catch (e) {
                response.writeHead(500, { "Content-Type": "application/json" });
                let msg = `Unable to process JSON POST body.\n${e.stack || e}\nPOST BODY from error: ${body}`;
                let retErr = this.makeAPIErrorResponse(500, typeof e == "string" ? e : e.message, msg);
                if (typeof e == "object" && e.meta) {
                    retErr.meta = e.meta;
                }
                response.end(JSON.stringify(retErr));
                return;
            }
        }
        // Document request
        request['transitRequestType'] = TransitRequestType.Document;
        let file;
        try {
            file = await getFile(this.config.contentDir, pathname);
        }
        catch (err) {
            try {
                file = await getFile(path.join(path.dirname(module.id), "../content"), pathname);
            }
            catch (err) {
                if (err.code == 'ENOENT') {
                    response.writeHead(404, { "Content-Type": "text/plain" });
                    response.end(`file not found: ${pathname}`);
                    response.end();
                }
                else {
                    response.writeHead(500, { "Content-Type": "text/plain" });
                    response.end(`Error reading file ${pathname}: ${err.code}`);
                    log.error(`Error reading file ${pathname}: ${err.code}`);
                    response.end();
                }
                return;
            }
        }
        // insert macros before checking for file staleness in case the inclusion
        // updates the file modification for a modified include file
        if (file.contentType == "text/html") {
            await this.insertMacros(this.config.contentDir, file, "<!--", "-->");
        }
        else if (file.contentType == "application/javascript") {
            await this.insertMacros(this.config.contentDir, file, "\\/\\*\\*", "\\*\\*\\/");
        }
        let ifModSince = Date.parse(request.headers["if-modified-since"]);
        if (isNaN(ifModSince) || Math.floor(ifModSince / 1000) < Math.floor(file.modified / 1000)) {
            response.writeHead(200, {
                'Content-Type': file.contentType,
                "Last-Modified": new Date(file.modified).toUTCString(),
                'Cache-Control': 'no-cache'
            });
            response.end(file.content, 'utf-8');
        }
        else {
            response.writeHead(304, {});
            response.end('', 'utf-8');
        }
        return;
    }
    async authorizedNamespaces(request) {
        let ident = null;
        if (this.identitySource != null) {
            let creds = await this.identitySource.obtainCredentials(request);
            if (creds) {
                ident = this.identitySource.validateCredentials(creds);
            }
        }
        return this.dataSourceFinder.allowedSources(ident);
    }
    async isAuthorized(request, response, ds) {
        // valid identity?
        if (this.identitySource != null) {
            let ident = null;
            let creds = await this.identitySource.obtainCredentials(request);
            let notifyCredentials = false;
            if (creds) {
                ident = this.identitySource.validateCredentials(creds);
            }
            if (!ident) {
                ident = await this.identitySource.obtainIdentity(request, response);
                notifyCredentials = true;
            }
            if (!ident || !ds.allowIdentity(ident)) {
                this.identitySource.configureFailureResponse(response);
                response.setHeader("Content-Type", "application/json");
                response.end(JSON.stringify({ errors: [{ status: 401, detail: "Sorry, you are not authorized to access this server." }] }));
                return null;
            }
            if (notifyCredentials) {
                response["authorization"] = this.identitySource.identityCredentials(ident);
            }
            return ident;
        }
        else {
            return { roles: [] };
        }
    }
    writeSuccessResponse(response, result) {
        response.writeHead(200, { "Content-Type": "application/json" });
        if (response["authorization"]) {
            // may have been filled in by Active Directory identity provider
            result.meta.authorization = response["authorization"];
        }
        response.end(JSON.stringify(result));
    }
    makeAPIErrorResponse(status, externalMessage, internalMessage) {
        if (status == 500) {
            log.error(internalMessage || externalMessage);
        }
        return { errors: [{ status: status, detail: externalMessage }] };
    }
    validateAPIRequest(ds, entity, operation, data) {
        var notFound = (message) => {
            return this.makeAPIErrorResponse(404, message);
        };
        // valid entity?
        if (!ds.hasEntity(entity))
            return notFound(`Entity ${ds.getName()}/${entity} not found.`);
        // operation supplied?
        if (operation == "")
            return notFound(`An operation must be specified for entity ${ds.getName()}/${entity}.`);
        // valid operation?
        if (!ds.hasOperation(entity, operation))
            return notFound(`Operation ${operation} not supported for entity ${ds.getName()}/${entity}.`);
        if (data === undefined) {
            data = [];
        }
        if (!Array.isArray(data)) {
            return this.makeAPIErrorResponse(500, `arguments must be an Array. POST body was:\n${JSON.stringify(data)}`);
        }
        return null;
    }
    getAPI(ns) {
        return this.dataSourceFinder.findSourceWithName(ns).api;
    }
    generateAPIJavaScript(ns) {
        let ds = this.dataSourceFinder.findSourceWithName(ns);
        let jsapi = this.apiJSTemplate;
        jsapi = jsapi.replace(/__API_NAME__/g, ns);
        jsapi = jsapi.replace(/__API_HIERARCHY__/g, JSON.stringify(ds.hierarchy));
        jsapi = jsapi.replace(/__CONSTRUCT_API__/g, ds.getAPIJavaScriptConstructor());
        jsapi = jsapi.replace(/__TRANSIT_API_VERSION__/g, String(TRANSIT_API_VERSION));
        return jsapi;
    }
    generateAPIHTML(ns) {
        let html = this.apiHTMLTemplate;
        html = html.replace(/__API_NAME__/g, ns);
        return html;
    }
    async executeSQL(dsName, sql, ...args) {
        sql = sql.replace(/\?+/g, (m) => {
            var ret = "?".repeat(m.length / 2);
            if (m.length % 2 && args.length) {
                var val = args.shift();
                switch (typeof val) {
                    case "number":
                        ret += isNaN(val) ? "NULL" : val;
                        break;
                    default:
                        ret += `'${String(val).replace(/'/g, "''")}'`;
                        break;
                }
            }
            return ret;
        });
        String((args.shift() || '')).replace(/'/g, "''");
        try {
            let result = await this.dataSourceFinder.findSourceWithName(dsName).executeSQL(sql);
            let rows = [];
            for (let row of result.rows) {
                let data = {};
                for (let i in result.columns) {
                    data[result.columns[i].spec.name] = row[i];
                }
                rows.push(data);
            }
            return rows;
        }
        catch (e) {
            throw new Error(`Error executing SQL: ${e}\nSQL statement: ${sql}`);
        }
    }
    async insertMacros(dir, file, openToken, closeToken) {
        let includes = {};
        let str = file.content.toString('utf-8');
        let re = `${openToken}INCLUDE:(.*?)${closeToken}`;
        let matches = str.match(new RegExp(re, "g"));
        if (matches) {
            for (let match of matches) {
                let submatch = match.match(new RegExp(re))[1];
                if (!includes[submatch]) {
                    let subfile = await getFile(dir + "/", submatch);
                    includes[submatch] = subfile.content.toString('utf-8');
                    file.modified = Math.max(file.modified, subfile.modified);
                }
            }
            str = str.replace(new RegExp(re, "g"), function (match) {
                return includes[match.match(new RegExp(re))[1]];
            });
        }
        str = str.replace(new RegExp(`${openToken}API_CONSTRUCTOR${closeToken}`, "g"), `Transit.constructAPIFromHierarchy = ${DataSource_1.constructAPIFromHierarchy.toString()}`);
        file.content = Buffer.from(str);
    }
}
exports.ServerRuntime = ServerRuntime;
function getFile(dir, pathname) {
    if (pathname == "/") {
        var filePath = dir + '/index.html';
    }
    else {
        var filePath = dir + pathname;
    }
    let extname = String(path.extname(filePath)).toLowerCase().substr(1);
    let contentType = Mimes_1.MimeTypes[extname] || 'application/octet-stream';
    return new Promise((resolve, reject) => {
        fs.stat(filePath, function (error, stats) {
            if (error) {
                reject(error);
            }
            else {
                fs.readFile(filePath, function (error, content) {
                    if (error) {
                        reject(error);
                    }
                    else {
                        resolve({
                            contentType: contentType,
                            modified: stats.mtime.getTime(),
                            content: content
                        });
                    }
                });
            }
        });
    });
}
function walkDir(dir, filelist = []) {
    fs.readdirSync(dir).forEach(file => {
        filelist = fs.statSync(path.join(dir, file)).isDirectory()
            ? walkDir(path.join(dir, file), filelist)
            : filelist.concat(path.join(dir, file));
    });
    return filelist;
}
async function readPostBody(request) {
    return new Promise(function (resolve, reject) {
        var body = '';
        request.on('data', function (data) {
            body += data;
            if (body.length > 100 * 1000 * 1000) {
                reject("POST body too big.");
            }
        });
        request.on('end', function () {
            resolve(body);
        });
    });
}
//# sourceMappingURL=Server.js.map