/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only. 
 */

import { CommandSpec, Command, EntitySpec, EntityType, DatabaseEntityBuilder, ParameterSpec, OutputSpec, DataType } from "./Command"
import { APISuccessResponse, DatabaseResult, DatabaseResultNotifier, OperationNotifier } from "./Result";
import { Identity } from "./IdentitySource";

declare var require : any;

export type DataSourceHierarchy = { 
	[key:string] : EntitySpec 
}

export type RequestProcessingFunction = (
	dsName : string, 
	entity : string, 
	operation : string, 
	args : any[]
) => Promise<APISuccessResponse>;

export type BatchRequestProcessingFunction = (
	dsName : string,
	args : [string, any[]][]
) => Promise<APISuccessResponse>;

export type ConstructAPIFunction = (
	hier : DataSourceHierarchy, 
	ns : string,
	processor : RequestProcessingFunction,
	batchProcessor : BatchRequestProcessingFunction
) => Object;

export interface DataSourceConfig {
	name: string;
	driver : string;
	driverConfig : DriverConfig;
	allowedUsernames?: string[];
	allowedRoles?: string[];
	extensions?: string[];
}

export enum HandlerType {
	Database = 0,
	Declared = 1
}

export interface Handler {
	readonly handlerType : HandlerType;
	loadEntitySpecs() : Promise<EntitySpec[]>;
	useAPI(api:any);
	executeCommand(command : Command) : Promise<APISuccessResponse>;
	executeBatch(commands : Command[]) : Promise<APISuccessResponse>;
}

export type DriverConfig = { 
	[key: string] : any 
};

export type DatabaseConnectionRef = any;

export interface Driver {
	readonly handlerType : HandlerType;
}

export interface DatabaseDriver extends Driver {
	openConnection(notifier : OperationNotifier<DatabaseConnectionRef>) : void;
	closeConnection(connection : DatabaseConnectionRef) : void;
	loadEntitySpecs(connection : DatabaseConnectionRef, builder : DatabaseEntityBuilder) : void;
	executeCommand(connection : DatabaseConnectionRef, command : Command, notifier: DatabaseResultNotifier) : void;
	executeSQL(connection : DatabaseConnectionRef, sql : string, notifier: DatabaseResultNotifier, outputs: OutputSpec[]|null) : void;
	beginTransaction(connection: DatabaseConnectionRef) : Promise<void>;
	commitTransaction(connection: DatabaseConnectionRef) : Promise<void>;
	rollbackTransaction(connection: DatabaseConnectionRef) : Promise <void>;
}

export interface DeclaredDriver extends Driver {
	useNamespaceAPI?(api:any),
	readonly API : { [key:string]: Function }
}

export class DataSourceFinder {
	readonly sources : { [key:string] : DataSource } = {};

	constructor() {
	}

	async addSource(
		dsconfig : DataSourceConfig, 
		processor : RequestProcessingFunction, 
		batchProcessor : BatchRequestProcessingFunction) 
	{
		try {
			var ds = new DataSource(dsconfig);
			await ds.constructAPI(dsconfig.name, processor, batchProcessor);
			this.sources[dsconfig.name] = ds;
		} catch(e) {
			let msg = `Error configuring ${dsconfig.name} DataSource with ${dsconfig.driver} driver: ${e}\n`;
			if (e instanceof Error) {
				e.message = msg + e.message;
				throw e;
			} else {
				throw msg + e;
			}
		}
	}

	hasSourceNamed(name : string) : boolean {
		return this.sources[name] != null;
	}
	
	findSourceWithName(name : string) : DataSource {
		if (this.sources[name]) {
			return this.sources[name];
		} else {
			throw new Error(`Data source ${name} not found.`);
		}
	}
	
	allowedSources(ident : Identity|null) : string[] {
		let allowed:string[] = [];
		for (var n in this.sources) {
			if (!ident || this.sources[n].allowIdentity(ident)) {
				allowed.push(n);
			}
		}
		return allowed;
	}
}

export class DataSource {
	private readonly config : DataSourceConfig;
	private readonly _entities: DataSourceHierarchy = {};
	private readonly handlers: { [key: string] : Handler } = {};
	private readonly extensions : Handler[] = [];

	protected handler : Handler;
	protected _api:any;
	
	get hierarchy() { return this._entities; }
	get api() { return this._api }

	constructor(dsconfig : DataSourceConfig) {
		this.config = dsconfig;

		let driverPath = dsconfig.driver.match(/^[\w\d]+$/) ? __dirname + "/driver/" + dsconfig.driver : dsconfig.driver;
		if (!dsconfig.name.match(/^\w[\w0-9_]*$/)) {
			throw new Error(`Namespace must begin with a letter and may contain only letters, numbers, or underscores.`);
		}
		this.handler = this.constructHandler(driverPath, dsconfig.driverConfig);

		for (let path of this.config.extensions||[]) {
			this.extensions.push(this.constructHandler(path, {}));
		}
	}

	constructHandler(driverPath:string, driverConfig:any) : Handler {
		let driver : Driver = (<any>require).main.require(driverPath).createDriver(driverConfig);
		switch (driver.handlerType) {
			case HandlerType.Database: return new DatabaseHandler(<DatabaseDriver>driver);
			case HandlerType.Declared: return new DeclaredHandler(<DeclaredDriver>driver);
			default:
				throw new Error(`The ${driverPath} driver provided an unknown handler type: ${driver.handlerType}`);
		}
	}

	getName() : string {
		return this.config.name;
	}

	allowIdentity(ident : Identity) : boolean {
		if (this.config.allowedUsernames 
			&& ident.username
			&& (this.config.allowedUsernames.find((val) => val == ident.username) || this.config.allowedUsernames.find((val) => val == "*"))
		) {
			return true;
		}

		if (this.config.allowedRoles) {
			for (let role of ident.roles) {
				if (this.config.allowedRoles.find((val) => val == role) || this.config.allowedRoles.find((val) => val == "*")) {
					return true;
				}
			}
		}

		return false;
	}

	hasEntity(entity: string) : boolean {
		try {
			this.specForEntity(entity);
			return true;
		} catch(e) {
			return false;
		}
	}

	hasOperation(entity : string, operation: string) : boolean {
		try {
			this.specForOperation(entity, operation);
			return true;
		} catch(e) {
			return false;
		}
	}

	defaultOperationName(entity : string) : string {
		let keys = Object.keys(this.specForEntity(entity).operations);
		if (keys.length == 1) {
			return keys[0];
		} else {
			return "";
		}
	}

	specForEntity(entityName : string) : EntitySpec {
		let entity  = this._entities[entityName];	
		if (!entity) {
			throw new Error(`Entity ${this.config.name}.${entityName} not found`);
		}
		return entity;
	}

	handlerForEntity(entityName : string) : Handler {
		let source  = this.handlers[entityName];
		if (!source) {
			throw new Error(`Entity data source ${this.config.name}.${entityName} not found`);
		}
		return source;
	}

	specForOperation(entityName : string, operation : string) : CommandSpec {
		let entity  = this.specForEntity(entityName);
		let opSpec = entity.operations[operation];
		if (!opSpec) {
			throw new Error(`${operation} operation not supported for entity ${this.config.name}.${entityName}`);
		}
		return { entity: entity, operation: opSpec };
	}

	async constructAPI(namespace:string, processor:RequestProcessingFunction, batchProcessor:BatchRequestProcessingFunction) {
		let entities = await this.handler.loadEntitySpecs();
		for (let entity of entities) {
			let apiName = Command.normalizeName(entity.name, this._entities);
			this._entities[apiName] = entity;
			this.handlers[apiName] = this.handler;
		}

		for (let ext of this.extensions) {
			let entities:EntitySpec[] = await ext.loadEntitySpecs();
			for (let entity of entities) {
				// extensions are allowed to override the the underlying entity of,
				// so pass in an empty env here.
				let apiName = Command.normalizeName(entity.name, {});
				this._entities[apiName] = entity;
				this.handlers[apiName] = ext;
			}
		}

		this._api = constructAPIFromHierarchy(this._entities, namespace, processor, batchProcessor);

		for (let handler of Object["values"](this.handlers)) {
			handler.useAPI(this.api);
		}
	}


	async executeCommand(entityName : string, operationName : string, args : any[]) : Promise<any> {
		let spec = this.specForOperation(entityName, operationName);
		let handler = this.handlerForEntity(entityName);
		if (spec == null) {
			throw new Error(`Object not found: ${this.config.name}.${entityName}`);
		}

		let command = new Command(spec, args || []);
		return await handler.executeCommand(command);
	}

	async executeBatch(args : [string, any[]][]) : Promise<APISuccessResponse> {
		let commands : Command[] = [];
		for (let i = 0; i < args.length; i++) {
			let [entityName, operationName] = args[i][0].split(".");
			let spec = this.specForOperation(entityName, operationName);
			if (this.handlerForEntity(entityName) != this.handler) {
				throw new Error("Batches may not contain APIs provided by extensions");
			}
			if (spec == null) {
				throw new Error(`Object not found: ${this.config.name}.${entityName}`);
			}
			commands.push(new Command(spec, args[i][1] || []));	
		}
		return await this.handler.executeBatch(commands);
	}

	async executeSQL(sql : string) : Promise<DatabaseResult> {
		if (this.handler.handlerType == HandlerType.Database) {
			return (this.handler as DatabaseHandler).executeSQL(sql);			
		} else {
			throw new Error(`Cannot call executeSQL with driver ${this.config.driver}`);
		}
	}

	getAPIJavaScriptConstructor() : string {
		return constructAPIFromHierarchy.toString();
	}
}

class DatabaseHandler implements Handler {
	readonly handlerType = HandlerType.Database;
	private driver : DatabaseDriver;

	constructor(driver : DatabaseDriver) {
		this.driver = driver;
	}
	
	async acquireConnection() : Promise<DatabaseConnectionRef> {
		let notifier = new OperationNotifier<DatabaseConnectionRef>();
		this.driver.openConnection(notifier);
		return await notifier.promise;
	}

	async releaseConnection(conn : DatabaseConnectionRef) {
		this.driver.closeConnection(conn);
	}

	async loadEntitySpecs() : Promise<EntitySpec[]> {
		let conn  = await this.acquireConnection();
		try {
			let builder = new DatabaseEntityBuilder();
			await this.driver.loadEntitySpecs(conn, builder);
			return await builder.promise;
		} finally {
			this.releaseConnection(conn);
		}
	}

	useAPI(api:any) {
	}

	async executeCommand(command : Command) : Promise<APISuccessResponse> {
		let conn = await this.acquireConnection();
		try {
			let notifier = new DatabaseResultNotifier();
			this.driver.executeCommand(conn, command, notifier);
			let response = await notifier.promise;
			return {
				data: response,
				meta: { handler: this.handlerType }
			};
		} finally {
			this.releaseConnection(conn);
		}
	}

	async executeBatch(commands : Command[]) : Promise<APISuccessResponse> {
		let results : DatabaseResult[] = [];
		let conn = await this.acquireConnection();
		let index = 0;
		try {
			await this.driver.beginTransaction(conn);
			for (index = 0; index < commands.length; index++) {		
				let notifier = new DatabaseResultNotifier();
				this.driver.executeCommand(conn, commands[index], notifier);
				results[results.length] = await notifier.promise;
			}
			await this.driver.commitTransaction(conn);
		} catch (err) {
			await this.driver.rollbackTransaction(conn);
			if (typeof err != "object") {
				err = new Error(err);
			}
			err.meta = { batchIndex: index };
			throw err;
		} finally {
			this.releaseConnection(conn);
		}
	
		return {
			data: results,
			meta: { handler: this.handlerType }
		};
	}

	async executeSQL(sql : string) : Promise<DatabaseResult> {
		let conn = await this.acquireConnection();
		try {
			let notifier = new DatabaseResultNotifier();
			this.driver.executeSQL(conn, sql, notifier, null);
			return await notifier.promise;
		} finally {
			this.releaseConnection(conn);
		}
	}	
}

class DeclaredHandler implements Handler {
	readonly handlerType = HandlerType.Declared;
	readonly driver : DeclaredDriver;

	constructor(driver : DeclaredDriver) {
		this.driver = driver;
	}

	async loadEntitySpecs() : Promise<EntitySpec[]> {
		let entities : EntitySpec[] = [];
		for (let spec in this.driver.API) {
			let regex = {
				identifier: /^([a-zA-Z_]\w*)/
			}

			var consumed = 0;
			var remaining = String(spec);

			function error(msg : string) {
				return new Error(`Unable to parse API signature - ${msg} at character ${consumed}: ${spec}`);
			}

			function whitespace() {
				if (remaining.match(/^(\s+)/)) {
					consumed += RegExp.$1.length;
					remaining = remaining.slice(RegExp.$1.length);
				}
			}
			function peekString(atom : string) : boolean {
				whitespace();
				return remaining.indexOf(atom) == 0;
			}

			function expectString(atom : string) : string {
				if (peekString(atom)) {
					consumed += atom.length;
					remaining = remaining.slice(atom.length);
					return atom;
				} else {
					throw error(`expected: ${atom}`);
				}
			}

			function peekMatch(re: string) {
				whitespace();
				return remaining.match(regex[re]);
			}

			function expectMatch(re : string) : string {
				if (peekMatch(re)) {
					consumed += RegExp.$1.length;
					remaining = remaining.slice(RegExp.$1.length);
					return RegExp.$1;
				} else {
					throw error(`expected ${re}`);
				}
			}

			function getType(type :string) : DataType {
				switch (type) {
					case "string": return DataType.String;
					case "float": return DataType.Float;
					case "int": return DataType.Integer;
					case "boolean": return DataType.Boolean;
					case "null": return DataType.Null;
					default: throw error(`unknown type ${type}`);
				}
			}

			let entity = expectMatch("identifier");

			expectString("(");
			let params : ParameterSpec[] = [];
			while (peekMatch("identifier")) {
				let name = expectMatch("identifier");
				expectString(":");
				let type = expectMatch("identifier");
				params.push({
					spec: {
						name: name,
						dataType: getType(type)
					},
					required: true,
					subparams: []
				});

				if (!peekString(",")) {
					break;
				}
			}
			expectString(")");
			
			expectString(":");
			
			let outputs : OutputSpec[] = [];
			let type = expectMatch("identifier");
			outputs.push({
				spec: {
					name: "value",
					dataType: getType(type)
				},
				writeable: false
			});

			entities.push({
				name: entity,
				type: 'function',
				identity: [],
				impl: this.driver.API[spec],
				operations: {
					call: {
						name: "call",
						parameters: params,
						outputs: outputs
					}
				}
			});
		}
		return Promise.resolve(entities);
	}

	useAPI(api:any) {
		if (this.driver.useNamespaceAPI) this.driver.useNamespaceAPI(api);
	}

	async executeCommand(command : Command) : Promise<APISuccessResponse> {
		let value = await command.spec.entity.impl.apply(this.driver.API, command.args);
		return  Promise.resolve({ 
			data: typeof value == 'undefined' ? null : value,
			meta: { handler: this.handlerType }
		});
	}

	async executeBatch(commands : Command[]) : Promise<APISuccessResponse> {
		let results : any[] = [];

		for (let index = 0; index < commands.length; index++) {
			let result;
			try {
				result = await this.executeCommand(commands[index]);
			} catch (err) {
				if (typeof err != "object") {
					err = new Error(err);
				}
				err.meta = { batchIndex: index };
				throw err;
			}
			results[results.length] = result.data;
		}	
		return {
			data: results,
			meta: { handler: this.handlerType }
		};
	}
}

// we're going to toString this for use in the browser,
// so we need to use old-school JS syntax...
export var constructAPIFromHierarchy : ConstructAPIFunction = function constructAPIFromHierarchy(
	hier : DataSourceHierarchy, 
	ns : string,
	processor : RequestProcessingFunction,
	batchProcessor : BatchRequestProcessingFunction
	) : Object 
{
	function remapRows(result : any) {
		var rows : any[] = [];
		for (var i = 0; i < result.rows.length; i++) {
			var row = result.rows[i];
			var data = {};
			for (var j = 0; j < result.columns.length; j++) {
				data[result.columns[j].spec.name] = row[j];
			}
			rows.push(data);
		}
		rows["columns"] = result.columns;
		return rows;
	}

	var ret = {
		_: { 
			NULL: 0, BOOLEAN: 1, INTEGER: 2, FLOAT: 3, STRING: 4, DATE: 5, DATETIME: 6, ARRAY: 7, OBJECT: 8, ANY: 9,		
		},
		$name: ns,
		"$batch_api": (function(d) {
			return function(args) {
				return new Promise((resolve, reject) => {
					batchProcessor(d, args)
					.then(function(response) {
						if (response.meta.handler == 0) {
							var remapped : any[] = [];
							for (var i = 0; i < response.data.length; i++) {
								remapped[i] = remapRows(response.data[i]);
							}
							resolve(remapped);
						} else {
							resolve(response.data);
						}
					})
					.catch(function(err) {
						reject(err)
					});
				});
			}
		})(ns)			
	};
	for (var entity in hier) {
		ret[entity] = {};
		for (var operation in hier[entity].operations) {
			(function(e, o) {
				ret[e][o] = function() {
					var args = Array.prototype.slice.call(arguments);
					return new Promise((resolve, reject) => {
						processor(ns, e, o, args)
						.then(function(response) {
							if (response.meta.handler == 0) {
								resolve(remapRows(response.data));
							} else {
								resolve(response.data);
							}
						})
						.catch(function(err) {
							reject(err);
						});
					});
				};
			})(entity, operation);
		}
		if (hier[entity].type == 'table') {
			(function(entName, entSpec, apiOut) {
				function unflattenRows(flat) {
					var headers = flat[0];
					var rows:any = [];

					var fieldSpecs = {};
					var outputs = entSpec.operations.select.outputs;
					for (var i = 0; i < outputs.length; i++) {
						fieldSpecs[outputs[i].spec.name.toLowerCase()] = outputs[i].spec;
					}

					for (var i = 1; i < flat.length; i++) {
						var row = {};
						for (var j = 0; j < flat[i].length; j++) {
							if (!headers[j]) continue;
							var fieldSpec = fieldSpecs[String(headers[j]).toLowerCase()];
							if (!fieldSpec) {
								throw new Error("Unknown column: " + headers[j]);
							}
							var value = flat[i][j];
							if (value != null) {
								switch (fieldSpec.dataType) {
									case ret._.BOOLEAN:
										if (typeof(value) == "string") {
											value = value.toLowerCase();
										}
										if (value == "n" || value == "no" || value == "f" || value == "false" || value == "0") {
											value = false;
										} else {
											value = !!value;
										}
										break;

									case ret._.INTEGER:
									case ret._.FLOAT:
										value = Number(value);
										if (isNaN(value)) {
											throw new Error("Value for column " + fieldSpec.name + " in row " + i + " is not a number");
										}
										break;

									case ret._.STRING:
										value = (value == null) ? null : String(value);
										break;

									default:
										// FALL THROUGH
								}
								row[fieldSpec.name] = value;
							}
						}
						rows.push(row);
					}
					return rows;
				}
				function buildRowBatch(rows) {
					var batch:any = [];
					for (var i = 0; i < rows.length; i++) {
						var params:any = [];
						var attrs = {};
						for (var n in rows[i]) {
							attrs[n] = rows[i][n];
						}
						for (var j = 0; j < entSpec.identity.length; j++) {
							var name = entSpec.identity[j].name;
							params.push(attrs[name]);
							delete attrs[name];
						}
						params.push(attrs);
						batch.push([entName + ".insert", params]);
					}
					return batch;
				}
				apiOut.buildInsertBatch = function(rows) {
					return buildRowBatch(rows);
				},
				apiOut.insertRows = function(rows) {
					return ret.$batch_api(this.buildInsertBatch(rows));
				}
				apiOut.insertRowsFlat = function(flat) {
					return this.insertRows(unflattenRows(flat));
				}
				apiOut.buildReplaceBatch = function(rows) {
					var batch = this.buildInsertBatch(rows);
					batch.unshift([entName + ".truncate"]);
					return batch;
				},
				apiOut.replaceRows = function(rows) {
					return ret.$batch_api(this.buildReplaceBatch(rows));
				}
				apiOut.replaceRowsFlat = function(flat) {
					return this.replaceRows(unflattenRows(flat));
				}
			})(entity, hier[entity], ret[entity]);
		}
	}
	return ret;
}

function keys(o : Object) : string[] {
	var k : string[] = [];
	for (let n in o) {
		k.push(n);
	}
	k.sort();
	return k;
}

