/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only. 
 */

import { OutputSpec } from './Command';
import { DatabaseConnectionRef, HandlerType } from "./DataSource";
import * as log from "./Logger";

export type Columns = OutputSpec[];
export type Row = any[];

export interface APISuccessResponse {
    data: any,
    meta: { handler: HandlerType, authorization?: string }
}

export interface APIErrorResponse {
    errors: APIError[],
    meta?: { batchIndex?: number }
}

export interface APIError {
    status: number,
    detail: string
}

export interface DatabaseResult {
    columns: Columns;
    rows: Row[];
}

export class OperationNotifier<T> {
    promise : Promise<T>;

    resultHandler   : (r : T)      => void;
    errorHandler    : (e : Error) => void;

    constructor() {
        let resolver : ((result:T) => void) | null = null;
        let rejecter : ((error:Error) => void) | null = null;
        this.promise = new Promise<T>((resolve, reject) => {
            resolver = resolve;
            rejecter = reject;
        });
    
        this.resultHandler   = (result : T)      => { resolver && resolver(result) };
        this.errorHandler    = (error : Error)   => { rejecter && rejecter(error) };
    }

    sendResult(r : T) {
        this.resultHandler(r);
        this._clearHandlers();
    }

    sendError(err : Error, debugInfo?: string) {
        log.error(`Execution failed with error:\n--- start of error\n${err.stack}\n--- end of error\n${debugInfo ? `Debug info: ${debugInfo}`:''}`);
        this.errorHandler(err);
        this._clearHandlers();
    }

    cancel() {
        this._clearHandlers();
    }

    _clearHandlers() {
        this.resultHandler   = (r) => {};
        this.errorHandler    = (e) => {};
    }
}

export class DatabaseResultNotifier extends OperationNotifier<DatabaseResult> {
    // todo: streaming results are commented out for now pending real-world need and automated tests. -NEM 3/15/18 

    resultHandler   : (r : DatabaseResult) => void;
    errorHandler    : (e : Error) => void;

    constructor()
    {
        super();
    }

    /* 
    sendColumns(c : Columns) {}
    sendRows(r : Row[]) {}
    sendComplete() {}
    */

    _clearHandlers() {
        super._clearHandlers();
    }
}
