/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only. 
 */

import * as ds from "../DataSource";
import * as res from "../Result";
import { DataType, EntitySpec, DatabaseEntityBuilder, Command, EntityType, OutputSpec } from "../Command";
import * as SQL from "../SQL";
import * as tds from "tedious";
import * as log from "../Logger";

export function createDriver(config) : ds.DatabaseDriver {
    return new Driver(config);
}

class Driver implements ds.DatabaseDriver {
    handlerType = ds.HandlerType.Database;
    config : any;

    constructor(config : any) {
        this.config = config;
    }

	openConnection(notifier : res.OperationNotifier<ds.DatabaseConnectionRef>) : void {
        var connection = new tds.Connection(this.config);
        connection.on('connect', function(err) {
            if (err) {
                notifier.sendError(err);
            } else {
                notifier.sendResult(connection);
            }
        });
        connection.on('error', function(err) {
            log.warn(`Database connection error: ${String(err)}`);
        });
    }

	closeConnection(connection : ds.DatabaseConnectionRef) : void {
        connection.close();
    }

	async loadEntitySpecs(connection:ds.DatabaseConnectionRef, builder:DatabaseEntityBuilder) {
        let sql = `
            SELECT *
            FROM (
                SELECT
                    'schema_name' = schema_name(procs.schema_id),
                    'entity_name' = object_name(procs.object_id),
                    'entity_type' = 'proc',
                    'col_name'    = SUBSTRING(params.name, 2, LEN(params.name)),
                    'data_type'   = type_name(user_type_id),  
                    'length'      = max_length,  
                    'ordinal'     = parameter_id,
                    'nullable'    = is_nullable,
                    'is_param'    = 1,
                    'required'    = IIf(is_nullable = 0, 1, 0),
                    'pk_index'    = 0,
                    'writeable'  = 0
                FROM sys.parameters params
                RIGHT JOIN sys.procedures procs ON params.object_id = procs.object_id
        
                UNION ALL
            
                SELECT
                    'schema_name' = schema_name(procs.schema_id),
                    'entity_name' = procs.name,
                    'entity_type' = 'proc',
                    'col_name'    = cols.name,  
                    'data_type'   = type_name(user_type_id),  
                    'length'      = max_length,  
                    'ordinal'     = column_ordinal,
                    'nullable'    = is_nullable,
                    'is_param'    = 0,
                    'required'    = 0,
                    'pk_index'    = 0,
                    'writeable'  = 0
                FROM sys.procedures procs
                cross apply (
                    SELECT * FROM sys.dm_exec_describe_first_result_set_for_object(object_id, 0)
                ) cols
                where error_state = 0
                
                UNION ALL
 
                SELECT
                    'schema_name' = schema_name(objs.schema_id),
                    'entity_name' = object_name(params.object_id),
                    'entity_type' = 'func',
                    'col_name'    = SUBSTRING(params.name, 2, LEN(params.name)),
                    'data_type'   = type_name(user_type_id),  
                    'length'      = max_length,  
                    'ordinal'     = parameter_id,
                    'nullable'    = is_nullable,
                    'is_param'    = 1,
                    'required'    = IIf(is_nullable = 0, 1, 0),
                    'pk_index'    = 0,
                    'writeable'  = 0
                FROM sys.parameters params
                INNER JOIN sys.objects objs ON params.object_id = objs.object_id
                WHERE objs.type='FN'
                AND is_output=0

                UNION ALL

                SELECT
                    'schema_name' = schema_name(objs.schema_id),
                    'entity_name' = object_name(params.object_id),
                    'entity_type' = 'func',
                    'col_name'    = SUBSTRING(params.name, 2, LEN(params.name)),
                    'data_type'   = type_name(user_type_id),  
                    'length'      = max_length,  
                    'ordinal'     = parameter_id,
                    'nullable'    = is_nullable,
                    'is_param'    = 0,
                    'required'    = 0,
                    'pk_index'    = 0,
                    'writeable'  = 0
                FROM sys.parameters params
                INNER JOIN sys.objects objs ON params.object_id = objs.object_id
                WHERE objs.type='FN'
                AND is_output=1
                
                UNION ALL
 
                SELECT
                    'schema_name' = schema_name(objs.schema_id),
                    'entity_name' = object_name(params.object_id),
                    'entity_type' = 'tfunc',
                    'col_name'    = SUBSTRING(params.name, 2, LEN(params.name)),
                    'data_type'   = type_name(user_type_id),  
                    'length'      = max_length,  
                    'ordinal'     = parameter_id,
                    'nullable'    = is_nullable,
                    'is_param'    = 1,
                    'required'    = IIf(is_nullable = 0, 1, 0),
                    'pk_index'    = 0,
                    'writeable'  = 0
                FROM sys.parameters params
                INNER JOIN sys.objects objs ON params.object_id = objs.object_id
                WHERE objs.type='TF'

                UNION ALL

                SELECT
                    'schema_name' = schema_name(objs.schema_id),
                    'entity_name' = objs.name,
                    'entity_type' = 'tfunc',
                    'col_name'    = cols.name,  
                    'data_type'   = type_name(user_type_id),  
                    'length'      = max_length,  
                    'ordinal'     = column_id,
                    'nullable'    = is_nullable,
                    'is_param'    = 0,
                    'required'    = 0,
                    'pk_index'    = 0,
                    'writeable'  = 0
                FROM sys.columns cols
                INNER JOIN sys.objects objs ON cols.object_id = objs.object_id
                WHERE objs.type='TF'

                UNION ALL
 
                SELECT
                    'schema_name' = schema_name(v.schema_id),
                    'entity_name' = v.name,
                    'entity_type' = 'view',
                    'col_name'    = cols.name,  
                    'data_type'   = type_name(user_type_id),  
                    'length'      = max_length,  
                    'ordinal'     = column_id,
                    'nullable'    = is_nullable,
                    'is_param'    = 0,
                    'required'    = 0,
                    'pk_index'    = 0,
                    'writeable'  = 0
                FROM sys.columns cols
                INNER JOIN sys.views v ON cols.object_id = v.object_id	

                UNION ALL

                SELECT
                    'schema_name' = schema_name(t.schema_id),
                    'entity_name' = t.name,
                    'entity_type' = 'table',
                    'col_name'    = cols.name,  
                    'data_type'   = type_name(user_type_id),  
                    'length'      = max_length,  
                    'ordinal'     = IIf(icols.key_ordinal > 0, icols.key_ordinal, 10000 + cols.column_id),
                    'nullable'    = is_nullable,
                    'is_param'    = 0,
                    'required'    = IIf(is_nullable = 0 AND default_object_id = 0 AND type_name(user_type_id) != 'timestamp', 1, 0),
                    'pk_index'    = ISNULL(icols.key_ordinal, 0),
                    'writeable'  = IIf(is_rowguidcol = 1 OR is_identity = 1 OR is_computed = 1, 0, 1)
                FROM sys.columns cols
                INNER JOIN sys.tables t ON cols.object_id = t.object_id	
                LEFT JOIN sys.index_columns icols ON cols.object_id = icols.object_id AND cols.column_id = icols.column_id
            ) data
            ORDER BY entity_name, ordinal
        `;
        
        let qryNotifier = new res.DatabaseResultNotifier();
        this.executeSQL(connection, sql, qryNotifier, null);
        let result = await qryNotifier.promise;
        let rows = result.rows.map((row) => ({ 
            schema_name: row[0],
            entity_name: row[1],
            entity_type: row[2],
            col_name:    row[3],
            data_type:   row[4],
            length:      row[5],
            ordinal:     row[6],
            nullable:    row[7],
            is_param:    row[8],
            required:    row[9],
            pk_index:    row[10],
            writeable:  row[11]
        }));

        let entTypeMap = {
            'proc'  : 'procedure',
            'func'  : 'procedure',
            'tfunc' : 'procedure',
            'view'  : 'view',
            'table' : 'table'
        };

        rows.forEach((row) => {
            builder.entity(
                row.entity_name,
                entTypeMap[row.entity_type],
                { schema: row.schema_name, type: row.entity_type });

            if (row.col_name == null) {
                // it's a proc that takes no params, we want to make sure it gets built...
                builder.proc();
                return;
            }

            let dataType = TypeMap[row.data_type];
            if (!dataType) {
                throw new Error(`SQL Server data type ${row.data_type} for ${row.entity_name}.${row.col_name} not yet supported`);
            }
            
            switch (entTypeMap[row.entity_type]) {
                case 'procedure':
                    if (row.is_param) {
                        builder.procParameter(row.col_name, dataType, row.data_type, !!row.required);
                    } else {
                        builder.procOutput(row.col_name, dataType, row.data_type);
                    }
                    break;

                case 'table':
                    builder.tableColumn(row.col_name, dataType, row.data_type, !!row.required, !!row.pk_index, row.writeable);
                    break;

                case 'view':
                    builder.viewColumn(row.col_name, dataType, row.data_type);
                    break;
            }
        }, null);

        builder.done();
    }

	executeCommand(connection : ds.DatabaseConnectionRef, command : Command, notifier: res.DatabaseResultNotifier) : void {
        let escaper = (value : string) => `[${value.replace(/\]/g, "]]")}]`.replace(/'/g, '');
        let namespacer = (entity : EntitySpec) => `${entity.impl.schema}.${escaper(entity.name)}`;

        let sql;
        switch (command.spec.entity.impl.type) {
            case 'proc':
                sql = `EXEC ${namespacer(command.spec.entity)} ${SQL.serializeSQLValueList(command.args)}`;
                break;
            case 'func':
                sql = `SELECT ${namespacer(command.spec.entity)}(${SQL.serializeSQLValueList(command.args)}) as result`;
                break;
            case 'tfunc':
                sql = `SELECT * FROM ${namespacer(command.spec.entity)}(${SQL.serializeSQLValueList(command.args)})`;
                break;
            case 'view':
                switch (command.spec.operation.name) {
                    case 'count':
                        sql = SQL.countSQL(command, namespacer, escaper);
                        break;
                    default:
                        sql = SQL.selectSQL(command, namespacer, escaper);
                        break;
                }
                break;
            default: // table
                switch (command.spec.operation.name) {
                    case 'insert':
                        sql = SQL.insertSQL(command, namespacer, escaper);
                        break;
                    case 'update':
                        sql = SQL.updateSQL(command, namespacer, escaper);
                        break;
                    case 'delete':
                        sql = SQL.deleteSQL(command, namespacer, escaper);
                        break;
                    case 'truncate':
                        sql = SQL.truncateSQL(command, namespacer, escaper);
                        break;
                    case 'count':
                        sql = SQL.countSQL(command, namespacer, escaper);
                        break;
                        default: // select
                        sql = SQL.selectSQL(command, namespacer, escaper);
                        break;
                } 
                break;  
        }

        this.executeSQL(connection, sql, notifier, command.spec.operation.outputs);
    }

	executeSQL(connection : ds.DatabaseConnectionRef, sql : string, notifier: res.DatabaseResultNotifier, outputs:OutputSpec[]|null) : void {
        let result = {
            columns: <res.Columns> [],
            rows: <any[][]> [],
        };
        let request = new tds.Request(sql, (err, rowCount) => {
            if (err) {
                notifier.sendError(err, sql);
            }
        });
        request.on("columnMetadata", columns => {
            result.columns = columns.map(c => { 
                let dataType = TypeMap[c.type.name];
                if (!dataType) {
                    notifier.sendError(new Error(`SQL Server data type ${c.type.name} for returned column ${c.colName} not yet supported`));
                    return null;
                } else {
                    let out = outputs ? outputs.find(o => o.spec.name == c.colName) : null;
                    let writeable = out ? out.writeable : false;
                    return { spec: { name: c.colName, dataType: dataType, implType: c.type.name }, writeable: writeable };
                }
            });
        });
        request.on("error", err => {
            notifier.sendError(err);
        });
        request.on("row", row => {
            result.rows.push(row.map(c => c.value));
        });
        request.on("requestCompleted", () => {
            notifier.sendResult(result);
        });
        connection.execSql(request);
    }

    beginTransaction(connection: ds.DatabaseConnectionRef) : Promise<void> {
        return new Promise(function(resolve, reject) {
            connection.beginTransaction((err) => { err ? reject(err) : resolve() });
        });
    }

	commitTransaction(connection: ds.DatabaseConnectionRef) : Promise<void> {
        return new Promise(function(resolve, reject) {
            connection.commitTransaction((err) => { err ? reject(err) : resolve() });
        });
    }

	rollbackTransaction(connection: ds.DatabaseConnectionRef) : Promise<void> {
        return new Promise(function(resolve, reject) {
            connection.rollbackTransaction((err) => { err ? reject(err) : resolve() });
        });
    }

}

var TypeMap = {
	Bit:  DataType.Boolean,
	BitN:  DataType.Boolean,

	TinyInt:    DataType.Integer,
	SmallInt:   DataType.Integer,
    Int:        DataType.Integer,
    IntN:       DataType.Integer,
	BigInt:     DataType.Integer,
	Numeric:    DataType.Float,
	NumericN:   DataType.Float,
	Decimal:    DataType.Float,
	DecimalN:   DataType.Float,
	SmallMoney: DataType.Float,
	Money:      DataType.Float,
	Float:      DataType.Float,
	MoneyN:     DataType.Float,
	FloatN:     DataType.Float,
	Real:       DataType.Float,

	SmallDateTime:   DataType.DateTime,
	DateTime:        DataType.DateTime,
	DateTimeN:       DataType.DateTime,
	DateTime2:       DataType.DateTime,
	DateTime2N:      DataType.DateTime,
	DateTimeOffset:  DataType.DateTime,
	DateTimeOffsetN: DataType.DateTime,
	Time:            DataType.DateTime,
	TimeN:           DataType.DateTime,
	Timestamp:       DataType.Integer, // aka rowversion in SQL Server
	Date:            DataType.Date,
	DateN:           DataType.Date,

	Char:      DataType.String,
	VarChar:   DataType.String,
	Text:      DataType.String,
	NChar:     DataType.String,
	NVarChar:  DataType.String,
	NText:     DataType.String,

	UniqueIdentifier:  DataType.String,
    UniqueIdentifierN: DataType.String,
    sysname:           DataType.String,
};

for (let n in TypeMap) {
    TypeMap[n.toLowerCase()] = TypeMap[n];
}

