/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only. 
 */

__API_NAME__ = (function() {
    function callAPI() {
        return api._.callAPI.apply(api._, arguments);
    }
    function callBatch(dsName, args) {
        return api._.callHTTP([dsName, "$batch-api"].join("/"),args);
    }

    var hier = __API_HIERARCHY__;
    var api = constructAPIFromHierarchy(hier, "__API_NAME__", callAPI, callBatch);

    api.$configureAuthorization = function(token, failCallback) {
        this._.authorization = token;
        this._.authFailureCallback = failCallback;
        var cont = this._.authContinuations;
        this._.authContinuations = [];
        for (var i = 0; i < cont.length; i++) {
            cont[i].call();
        }
    }

    api.$authorize = function() {
        return api._.callHTTP(["__API_NAME__", "$authorize"].join("/"), {});
    }

    api._.authorization = null;
    api._.authFailureCallback = null;
    api._.authContinuations = [];
    api._.callAPI = function(dsName, entity, operation, args) {
        return this.callHTTP([dsName, entity, operation].join("/"), args);
    };
    api._.callHTTP = function(path, args) {
        return new Promise(function(resolve, reject) {
            var oXHR = new XMLHttpRequest();
            oXHR.open("POST", (api.$baseURL || "") + "/v__TRANSIT_API_VERSION__/" + path, true);
            oXHR.setRequestHeader("Content-Type", "application/json;charset=UTF-8");
            var auth = api._.authorization;
            if (typeof auth == "function") {
                auth = auth();
            }
            if (typeof auth == "string") {
                oXHR.setRequestHeader("x-transit-authorization", auth);
            }
            oXHR.onreadystatechange = function (oEvent) { 
                if (oXHR.readyState === 4) {  
                    if (oXHR.status === 200) {
                        try {
                            var json = JSON.parse(oXHR.responseText);
                            if (json.meta.authorization) {
                                api._.authorization = json.meta.authorization;
                            }
                            resolve(json);
                        } catch(e) {
                            reject(e.toString() + " " + oXHR.responseText);
                        }
                    } else if (oXHR.status == 401 && api._.authFailureCallback) {
                        var notifyFail = api._.authContinuations.length == 0;
                        api._.authContinuations.push(function() {
                            api._.callHTTP(path, args)
                            .then(function(result) {
                                resolve(result);
                            }).catch(function(err) {
                                reject(err);
                            });
                        });
                        if (notifyFail) {
                            api._.authFailureCallback.call();
                        }
                    } else {
                        try {
                            reject(JSON.parse(oXHR.responseText).errors[0].detail);
                        } catch(e) {
                            if (oXHR.statusText == "" && oXHR.responseText == "") {
                                reject("Network request failed.");
                            } else {
                                reject(oXHR.statusText + " - " + oXHR.responseText);
                            }
                        }
                    }  
                }  
            }; 
            oXHR.send(JSON.stringify(args));
        });
    };
    api._.hierarchy = hier;

    for (var entity in hier) {
        for (var operation in hier[entity].operations) {
            api[entity][operation].vue = makeMixinFactory(entity, operation);
        }
        if (hier[entity].type == 'table') {
            api[entity].vue = makeMixinFactory(entity, null);
        }
    }

    api.vueSave = function(/*...*/) {
        var objects = [].slice.call(arguments);
        var batch = this.vueBuildBatch(objects);
        return new Promise(function(resolve, reject) {
            api.$batch_api(batch)
            .then(function(result) {
                for (var i = 0; i < objects.length; i++) {
                    var vue = objects[i];
                    if (vue.$transit) {
                        // it's an individual row
                        vue.$transit._afterSave();
                    } else {
                        for (var j = 0; j < vue.rows.length; j++) {
                            vue.rows[j].$transit._afterSave();
                        }
                    }
                }
                resolve(result);
            }).catch(function(err) {
                reject(err);
            });
        });
    }

    // separate function allows for monkey-patching
    api.vueBuildBatch = function(objects) {
        var batch = [];
        for (var i = 0; i < objects.length; i++) {
            var vue = objects[i];
            if (vue.$transit) {
                // it's an individual row
                batch.push(vue.$transit._buildSaveSpec());
            } else {
                for (var j = 0; j < vue.rows.length; j++) {
                    batch.push(vue.rows[j].$transit._buildSaveSpec());
                }
                for (var j = vue.removed.length - 1; j >= 0; j--) {
                    batch.push(vue.removed[j].$transit._buildDeleteSpec());
                    vue.removed[j].$transit._detach();
                }
            }
        }
        return batch;
    }

    var keySequence = 0;
    function makeMixinFactory(argEntity, argOperation) {
        return function(propName, defaultArgs) {
            // cache our factory args locally so that if we change them,
            // we don't impact other factories made by this generator.
            var entity = argEntity;
            var operation = argOperation;

            var enableEditing = false;
            if (operation == null) {
                operation = "select";
                enableEditing = true;
            }
            var params = hier[entity].operations[operation].parameters;
            var identity = hier[entity].identity;
            propName = propName || entity;
            var $root;
            var mixin = {
                created: function() {
                    $root = this;
                },
                data: {},
                watch: {},
                load: function() {
                    mdata.execute();
                    return this;
                },
                onExecute: function(callback) {
                    mdata.onExecute = callback;
                    return this;
                },
                onError: function(callback) {
                    mdata.onError = callback;
                    return this;
                }
            }
            var emptyRows = [];
            emptyRows.columns = [];
            var mdata = mixin.data[propName || entity] =  {
                args: defaultArgs || {},
                rows: patchRows(emptyRows),
                row: {},
                removed: [],
                result: undefined,
                unused: true,
                inProgress: false,
                complete: false,
                error: "",
                onExecute: null,
                onError: null,
                execute: function() { return execute(true) },
                csv: function() {
                    function escapeCSV(value) {
                        value = (value == null || value == undefined) ? "" : String(value);
                        if (value.match(/[,\n]/)) {
                            value = value.replace(/"/g, '""');
                            value = '"' + value + '"';
                        }
                        return value;
                    }
                    
                    var outrows = [];
                    var row = [];
                    var cols = this.rows.columns;
                    for (var i = 0; i < cols.length; i++) {
                        row.push(escapeCSV(cols[i].spec.name));
                    }
                    outrows.push(row.join());
                    for (var i = 0; i < this.rows.length; i++) {
                        row = [];
                        for (var j = 0; j < cols.length; j++) {
                            row.push(escapeCSV(this.rows[i][cols[j].spec.name]));
                        }
                        outrows.push(row.join());
                    }
                    outrows.push(""); // adds trailing newline
                    return outrows.join("\n");
                }
            };

            var dirty = false;
            var iteration = 0;

            function makeIdentity(row) {
                var ident = [];
                for (var i = 0; i < identity.length; i++) {
                    ident.push(row[identity[i].name]);
                }
                ident.toString = ident.join;
                return ident;
            }
            function patchRow(row, rows) {
                var successFn = null, errorFn = null, unwatch = null;
                Vue.set(row, "$transit",  {
                    key: keySequence++,
                    storedIdentity: makeIdentity(row), // note that this is the ORIGINAL identity for the row,
                                                       // needs to update if the identity changes.
                    isUnique: function() {
                        var key = makeIdentity(row).toString();
                        for (var i = 0; i < rows.length; i++) {
                            if (rows[i] != row && rows[i].$transit.storedIdentity.toString() == key) {
                                return false;
                            }
                        }
                        return true;
                    },
                    edit: function() {
                        // store original values
                        this.previous = this._copy();
                        this.editing = true;
                        if (this.saving) {
                            console.log("transit warning: calling edit while saving has undefined behavior");
                        }
                    },
                    save: function() {
                        if (this.saving) {
                            throw new Error("transit error: save attempted while already saving.");
                        }

                        var operation = this._saveOperation();
                        var args = this._saveArgs();
                        this.saving = true;
                        return new Promise(function(resolve, reject) {
                            if (row.$transit) {
                                row.$transit.editing = false;
                                api[entity][operation].apply(null, args)
                                .then(function() {
                                    if (row.$transit) {
                                        row.$transit._afterSave();
                                    }
                                    resolve();
                                    if (successFn) successFn(row);
                                })["catch"](function(err) {
                                    if (row.$transit) {
                                        row.$transit.editing = true;
                                        row.$transit.saving = false;
                                    }
                                    reject(err);
                                    if (errorFn) errorFn(err, row);
                                });
                            }
                        });
                    },
                    autosave: function(success, error) {
                        successFn = success;
                        errorFn = error;
                        var enabled = true;
                        unwatch = $root.$watch(function() { return row; }, function(val, old) {
                            // only save while not in the act of saving, otherwise the save
                            // triggers a new update and you get stuck in an infinite loop.
                            // Don't rely on the saving boolean because its update is triggered
                            // asynchronously.
                            if (enabled) {
                                enabled = false;
                                row.$transit.save().then(function() {
                                    $root.$nextTick(function() {
                                        enabled = true;
                                    });
                                }).catch(function() {
                                    $root.$nextTick(function() {
                                        enabled = true;
                                    });
                                });
                            }
                        }, { deep: true });
                    },
                    cancel: function() {
                        if (this.editing) {
                            for (var n in this.previous) {
                                row[n] = this.previous[n];
                            }
                            delete this.previous;
                            this.editing = false;
                        }
                    },
                    remove: function() {
                        for (var i = 0; i < rows.length; i++) {
                            if (rows[i] == row) {
                                rows.splice(i, 1);
                                break;
                            }
                        }
                    },
                    removeAndSave: function() {
                        this.remove();
                        var $transit = this;
                        this._detach();
                        if (!this.isNew) {
                            return new Promise(function(resolve, reject) {
                                api[entity].delete.apply(null, $transit.storedIdentity)
                                .then(function() {
                                    resolve();
                                })["catch"](function(err) {
                                    reject(err, row);
                                });
                            });
                        } else {
                            return Promise.resolve();
                        }
                    },
                    _saveOperation: function() {
                        return this.isNew ? "insert" : "update"
                    },
                    _saveArgs: function() {
                        var values = this._copy();
                        var opName = this._saveOperation();
                        if (opName == "insert") {
                            // remove duplicate primary key values
                            for (var i = 0; i < identity.length; i++) {
                                delete values[identity[i].name];
                            }
                        } else { // update
                            // remove values that aren't writeable
                            for (var i = 0; i < rows.columns.length; i++) {
                                if (!rows.columns[i].writeable) {
                                    delete values[rows.columns[i].spec.name];
                                }
                            }
                        }
                        // on update we need to use the original primary keys, but on insert
                        // we need to use the current primary keys
                        var apiIdent = (opName == "insert" ? makeIdentity(row) : this.storedIdentity);
                        return apiIdent.concat(values);
                    },
                    _afterSave: function() {
                        this.isNew = false;
                        this.saving = false;
                        delete this.previous;
                        // our identity may have changed during edits...
                        this.storedIdentity = makeIdentity(row);
                    },
                    _buildSaveSpec: function() {
                        return [entity + "." + this._saveOperation(), this._saveArgs()];
                    },
                    _buildDeleteSpec: function() {
                        if (this.isNew) {
                            throw new Error("Transit error: Can't build delete command for a row that was never inserted");
                        }
                        return [entity + ".delete", this.storedIdentity.concat()];
                    },
                    _detach: function() {
                        if (unwatch) unwatch();
                        for (var i = 0; i < mdata.removed.length; i++) {
                            if (mdata.removed[i] == row) {
                                mdata.removed.splice(i,1);
                                break;
                            }
                        }       
                        row.$transit = null;
                    },
                    _copy: function() {
                        var c = {};
                        for (var n in row) {
                            if (n != "$transit") c[n] = row[n];
                        }
                        return c;
                    },
                    isNew: false,
                    editing: false,
                    saving: false
                });

                return row;
            }

            function stashRemoved(row) {
                if (!row.$transit.isNew) {
                    mdata.removed.push(row);
                }
            }
            function patchRows(rows) {
                if (!enableEditing || !identity.length) return rows;

                var orig = {
                    push: rows.push,
                    pop: rows.pop,
                    shift: rows.shift,
                    unshift: rows.unshift,
                    splice: rows.splice
                };

                function patchRowSet(rowSet, start) {
                    start = start || 0;
                    for (var i = start; i < rowSet.length; i++) {
                        if (typeof rowSet[i] != "object" || rowSet[i] == null) {
                            throw new Error("transit row element must be an object");
                        }
                        rowSet[i] = patchRow(rowSet[i], rows);
                        rowSet[i].$transit.isNew = true;
                    }
                    return rowSet;
                }

                rows.push = function() {
                    patchRowSet(arguments);
                    return orig.push.apply(this, arguments);
                }
                rows.pop = function() {
                    if (rows.length) {
                        stashRemoved(rows[rows.length-1]);
                    }
                    return orig.pop.apply(this, arguments);
                }
                rows.unshift = function() {
                    patchRowSet(arguments);
                    return orig.unshift.apply(this, arguments);
                }
                rows.shift = function() {
                    if (rows.length) {
                        stashRemoved(rows[0]);
                    }
                    return orig.shift.apply(this, arguments);
                }
                rows.splice = function(start, length) {
                    var removed = rows.slice(start, start + length);
                    for (var i = 0; i < removed.length; i++) {
                        stashRemoved(removed[i]);
                    }
                    patchRowSet(arguments, 2);
                    return orig.splice.apply(this, arguments);
                }

                for (var i = 0; i < rows.length; i++) {
                    patchRow(rows[i], rows);
                }
                return rows;
            }

            function execute(returnPromise) {
                iteration++;
                dirty = true;
                mdata.unused = false;
                mdata.inProgress = true;
                mdata.complete = false;
                mdata.error = "";
                var resolve, reject, promise;
                if (returnPromise) {
                    promise = new Promise(function(_resolve, _reject) {
                        resolve = _resolve;
                        reject = _reject;
                    });
                }
                Vue.nextTick(function() {
                    if (dirty) {
                        var args = [];
                        for (var i = 0; i < params.length; i++) {
                            var val = mdata.args[params[i].spec.name];
                            if (typeof val == "function") {
                                val = val.call();
                            }
                            args.push(val);
                        }
                        dirty = false;
                        var callIteration = iteration;
                        api[entity][operation].apply(api[entity], args)
                        .then(function(result) {
                            if (iteration == callIteration) {
                                if (Array.isArray(result)) {
                                    mdata.rows = result;
                                    mdata.row = result[0] || {};
                                } else {
                                    mdata.result = result;
                                }
                                mdata.inProgress = false;
                                mdata.complete = true;
                                if (mdata.onExecute) {
                                    mdata.onExecute();
                                }
                                if (resolve) resolve(mdata.rows);
                            }
                        })["catch"](function(err) {
                            if (iteration == callIteration) {
                                mdata.inProgress = false;
                                mdata.complete = true;
                                mdata.error = String(err);
                                if (mdata.onError) {
                                    mdata.onError(mdata.error);
                                }
                                if (reject) reject(mdata.error);
                            }
                        });
                    }
                });
                return promise;
            }

            for (var i = 0; i < params.length; i++) {
                var pname = params[i].spec.name;
                // make our properties reactive
                if (mdata.args[pname] === undefined) {
                    mdata.args[pname] = undefined;
                }
            }
            mixin.watch[propName+".args"] = {
                handler: function() { execute(false) },
                deep: true
            };

            mixin.watch[propName+".rows"] = function(val, prev) {
                if (val != prev) {
                    patchRows(val);
                }
            };
            
            return mixin;
        }
    }

    return api;

    __CONSTRUCT_API__

})();