declare type APISuccessResponse   = import("../../../lib/Result").APISuccessResponse;
declare type APIErrorResponse     = import("../../../lib/Result").APIErrorResponse;
declare type DatabaseResult       = import("../../../lib/Result").DatabaseResult;
declare type DataSourceHierarchy  = import("../../../lib/DataSource").DataSourceHierarchy;
declare type ConstructAPIFunction = import("../../../lib/DataSource").ConstructAPIFunction;
declare type EntitySpec           = import("../../../lib/Command").EntitySpec;
declare type ParameterSpec        = import("../../../lib/Command").ParameterSpec;
declare type EntityType           = import("../../../lib/Command").EntityType;

declare type RequestProcessingFunction      = import("../../../lib/DataSource").RequestProcessingFunction;
declare type BatchRequestProcessingFunction = import("../../../lib/DataSource").BatchRequestProcessingFunction;

namespace Transit {
    type AuthToken = string;
    type EntityRef = string;
    type EntityOptions = {
        entity?: string,
        wait?: boolean,
        single?: boolean,
        editable?: boolean,
        autosave?: boolean,
        args?: EntityArgs,
        onload?: Function,
        onsave?: Function,
        onloaderror?: Function,
        onsaveerror?: Function,
    }
    type EntityStatus = {
        unused: boolean,
        edited: boolean,
        loading: boolean,
        saving: boolean,
        complete: boolean,
        error: string,
        pageCount: number,
    }
    type EntityArgs = { [key:string]: any };
    type Row = { [key:string]: any };

    type BatchEntry = [string, Object];
    type Batch = BatchEntry[];

    type RequestSpec = {ref:EntityRef, batchCount:number, version?: number, resolve?:Function, reject?:Function};

    type Editable = {
        [key: string] : any,
        $previous: null|Editable,
        $ref: EntityRef,
        $new: boolean,
        $alive: boolean,
    };

    export function loadNamespaceList(authToken?:string) {
        return new Promise(async (resolve, reject) => {
            let response = await Transit.callServer(`namespaces`, undefined, authToken);
            if (Transit.isSuccess(response)) {
                resolve(response.data);
            } else {
                reject(new Error(response.errors[0].detail));
            }
        });
    }

    export function loadNamespace(ns:string) : Promise<Namespace> {
        return new Promise(async (resolve, reject) => {
            let response = await Transit.callServer(`${ns}.json`);
            if (Transit.isSuccess(response)) {
                resolve(this['$'+ns] = new Namespace(ns, response.data));
            } else {
                reject(new Error(response.errors[0].detail));
            }
        });
    }

    class Namespace {
        name:string;
        hierarchy:DataSourceHierarchy;
        api:any;
        authToken:AuthToken;
        entities:{[key:string]:Entity} = {};
        loadRequests:RequestSpec[] = [];
        vue:any = null;

        initResolve:Function|null;
        initReject:Function|null;
        oninitPromise:Promise<void>;
        onautherror:Function;

        constructor(name:string, hierarchy:DataSourceHierarchy) {
            this.name = name;
            this.hierarchy = hierarchy;
            let apiConstructor:ConstructAPIFunction = Transit["constructAPIFromHierarchy"];
            this.api = apiConstructor(
                hierarchy, 
                name, 
                (dsName, entity, operation, args) => {
                    // TODO: supply auth
                    return new Promise<APISuccessResponse>((resolve, reject) => {
                        this.callServerAuthed(`${dsName}/${entity}/${operation}`, args)
                        .then(function(response) {
                            if (isSuccess(response)) {
                                resolve(response);
                            } else {
                                reject(response.errors[0].detail);
                            }
                        })
                    });
                },
                (dsname, args) => {
                    return new Promise<APISuccessResponse>((resolve, reject) => {
                        this.callServerAuthed(`${dsname}/$batch-api`, args)
                        .then(function(response) {
                            if (isSuccess(response)) {
                                resolve(response);
                            } else {
                                reject(response.errors[0].detail);
                            }
                        })
                    });
                }
            );
            this.oninitPromise = new Promise<void>((resolve, reject) => {
                this.initResolve = resolve;
                this.initReject = reject;
            });
        }

        makeEntity(ref:string, options:EntityOptions) : Entity {
            if (this.entities[ref]) {
                throw new Error("Transit: entity ${ref} already exists");
            }
            if (!options.entity) {
                options.entity = ref;
            }
            if (this.hierarchy && !this.hierarchy[options.entity]) {
                throw new Error("Transit: entity not found: " + options.entity);
            } else {
                return this['$'+ref] = this.entities[ref] = new Entity(ref, this, this.hierarchy[options.entity], options||{});
            }
        }

        vueMixin(specs:{[key:string]:EntityOptions}) : any {
            let data:any = {};
            let watch:any = {};
            let computed:any = {};
            for (let ref in specs) {
                let entity = this.makeEntity(ref, specs[ref]);
                data[ref] = entity.result();
                data[`${ref}Status`] = entity.status;
                if (typeof entity.args == "function") {
                    computed[`${ref}Args`] = entity.args;
                } else {
                    data[`${ref}Args`] = entity.args;
                }
                watch[`${ref}Args`] = {
                    handler: ((ref) => {
                        return () => this.scheduleLoadRequest(ref, false);
                    })(ref),
                    deep: true
                }
                if (!entity.options.wait && entity.spec.type != 'procedure' && entity.spec.type != 'function') {
                    this.scheduleLoadRequest(ref, false);
                }
            }

            if (!this.loadRequests.length) {
                // we have no load requests -- we're initialized
                setTimeout(() => {
                    if (this.initResolve) {
                        this.initResolve();
                        this.initResolve = null;
                        this.initReject = null;
                    }
                }, 1);
            }

            let _ns = this;
            return {
                data: data,
                computed: computed,
                watch: watch,
                created: function() {
                    _ns.vue = this;
                    for (let ref in specs) {
                        // need to do this after Vue has instrumented the result so that
                        // our custom observation methods aren't bypassed.
                        _ns.entities[ref].configureResult();
                    }
                }
            };
        }

        findRef(ref:EntityRef|Editable[]) : EntityRef {
            if (typeof ref != "string") {
                ref = (<any>ref).$ref;
                if (!ref) {
                    throw new Error("Not a transit entity");
                }
            } else {
                if (!ref) {
                    throw new Error("Entity reference not found: " + ref);
                }
            }
            return <EntityRef>ref;
        }

        load(ref:EntityRef|Editable[]) : Promise<any> {
            return this.scheduleLoadRequest(this.findRef(ref), true)!;
        }

        save(...refs:(EntityRef|Editable[])[]) : Promise<any> {
            for (let i = 0; i < refs.length; i++) {
                refs[i] = this.findRef(refs[i]);
            }
            return this.processSaveRequest(<EntityRef[]>refs); // don't await
        }

        async revert(...refs:(EntityRef|Editable[])[]) {
            for (let ref of refs) {
                await this.entities[this.findRef(ref)].revert();
            }
        }

        truncate(...refs:(EntityRef|Editable[])[]) : Promise<any> {
            for (let i = 0; i < refs.length; i++) {
                refs[i] = this.findRef(refs[i]);
            }
            return this.processTruncateRequest(<EntityRef[]>refs); // don't await
        }
            
        isUnique(row:Row) : boolean {
            return this.entities[row.$ref].editedRowIdentityCount(row) <= 1;
        }

        csv(ref:EntityRef|Editable[]) {
            return this.entities[this.findRef(ref)].csv();
        }

        count(ref:EntityRef|Editable[]) {
            return this.entities[this.findRef(ref)].rows.length;
        }

        scheduleLoadRequest(ref:EntityRef, returnPromise:boolean) : Promise<any>|null {
            let entity = this.entities[ref];
            if (!entity) {
                throw new Error("Transit: entity not found: " + ref);
            }
            entity.loadRequested = true;
            entity.loadRequestSequence++;
            entity.status.unused = false;
            entity.status.loading = true;
            entity.status.saving = false;
            entity.status.complete = false;
            entity.status.error = "";
            let entry:RequestSpec = {ref: ref, batchCount: 1};
            this.loadRequests.push(entry);
            this.processLoadRequests();
            if (returnPromise) {
                return new Promise((resolve, reject) => {
                    entry.resolve = resolve;
                    entry.reject  = reject;
                });
            } else {
                return null;
            }
        }

        async processLoadRequests() {
            // don't schedule multiple requests,
            // and make sure to bundle all requests in an execution
            // cycle into one batch.
            await delay(1);
            if (this.loadRequests.length == 0) {
                return;
            }

            let requests:RequestSpec[] = [];
            let batch:Batch = [];
            for (let request of this.loadRequests) {
                let entity = this.entities[request.ref];
                request.version = entity.loadRequestSequence;
                if (entity.loadRequested) {
                    let items = entity.makeLoadBatchItems(this.vue[`${request.ref}Args`]);
                    batch = batch.concat(items);
                    // note how many results need to be consumed when processing results.
                    // (Used for updating pageCount with paged results.)
                    request.batchCount = items.length;
                    requests.push(request);
                    entity.loadRequested = false;
                }
            }
            this.loadRequests = [];

            let response = await this.callServerAuthed(`${this.name}/$batch-api`, batch);
            let responseIndex = 0;
            for (let i = 0; i < requests.length; i++) {
                let request = requests[i];
                let entity = this.entities[request.ref];

                if (request.version != entity.loadRequestSequence) {
                    // our entity state changed while the request was in-transit,
                    // ignore this response
                } else {
                    entity.status.loading = false;
                    entity.status.complete = true;
                    entity.updating = true;
                    if (isSuccess(response)) {           
                        let returned = entity.processLoadResponseItems(response.data.slice(responseIndex, responseIndex + request.batchCount));
                        entity.status.error = "";
                        if (entity.onload) entity.onload!.call(this.vue, returned);
                        if (request.resolve) request.resolve!(returned);
                    } else {
                        if (!response.meta || typeof response.meta.batchIndex != "number" || i == response.meta.batchIndex) {
                            let errmsg = entity.status.error = response.errors[0].detail;
                            if (entity.onloaderror) entity.onloaderror!.call(this.vue, errmsg);
                            if (request.reject) request.reject!(new Error(errmsg));
                        }                        
                    }
                }
                responseIndex += request.batchCount;
            }

            for (let i = 0; i < requests.length; i++) {
                this.entities[requests[i].ref].updating = false;
            }

            if (this.initResolve) {
                if (isSuccess(response)) {
                    this.initResolve!();
                } else {
                    this.initReject!(new Error(response.errors[0].detail));
                }
                this.initResolve = null;
                this.initReject = null;
            }
        }

        async processSaveRequest(refs:EntityRef[]) : Promise<void> {
            let batch:Batch = [];
            let rows:Editable[] = [];
            for (let ref of refs) {
                let entity = this.entities[ref];

                if (!entity.options.editable) {
                    throw new Error(`Cannot save entity '${entity.ref}' because it is not configured to be editable`);
                }

                for (let row of entity.resultAsRows()) {
                    if (row.$new) {
                        batch.push(entity.makeInsertBatchItem(<Editable>row));
                        rows.push(<Editable>row);
                    } else if (row.$previous) {
                        batch.push(entity.makeUpdateBatchItem(<Editable>row));
                        rows.push(<Editable>row);
                    }
                }

                for (let row of entity.removed) {
                    batch.push(entity.makeDeleteBatchItem(row));
                    rows.push(row);
                }

                entity.status.unused = false;
                entity.status.saving = true;
                entity.status.complete = false;
                entity.status.error = "";
            }

            if (!batch.length) {
                for (let ref of refs) {
                    let entity = this.entities[ref];
                    entity.status.saving = false;
                    entity.status.edited = false;
                    entity.status.complete = true;
                    if (entity.onsave) entity.onsave!.call(this.vue, entity.result());
                }
                return;
            }
            
            let response = await this.callServerAuthed(`${this.name}/$batch-api`, batch);
            for (let row of rows) {
                if (isSuccess(response)) {
                    row.$previous = null;
                    row.$new = false;
                }
            }

            for (let ref of refs) {
                let entity = this.entities[ref];
                entity.status.saving = false;
                entity.status.complete = true;
                if (isSuccess(response)) {
                    entity.status.edited = false;
                    entity.status.error = "";
                    entity.markRevertPoint();
                    if (entity.onsave) entity.onsave!.call(this.vue, entity.result());
                } else {
                    entity.status.error = response.errors[0].detail;
                    if (entity.onsaveerror) entity.onsaveerror!.call(this.vue, entity.status.error);
                }
            }

            if (!isSuccess(response)) {
                throw new Error(response.errors[0].detail);
            }
        }

        async processTruncateRequest(refs:EntityRef[]) : Promise<void> {
            let batch:Batch = [];
            for (let ref of refs) {
                let entity = this.entities[ref];

                if (!entity.options.editable) {
                    throw new Error(`Cannot save entity '${entity.ref}' because it is not configured to be editable`);
                }

                batch.push(entity.makeTruncateBatchItem());

                entity.status.unused = false;
                entity.status.saving = true;
                entity.status.complete = false;
                entity.status.error = "";
            }

            let response = await this.callServerAuthed(`${this.name}/$batch-api`, batch);
            for (let ref of refs) {
                let entity = this.entities[ref];
                entity.status.saving = false;
                entity.status.complete = true;
                entity.updating = true;
                if (isSuccess(response)) {
                    entity.replaceResult([]);
                    entity.status.error = "";
                    entity.markRevertPoint();
                    if (entity.onsave) entity.onsave!.call(this.vue, entity.result());
                } else {
                    entity.status.error = response.errors[0].detail;
                    if (entity.onsaveerror) entity.onsaveerror!.call(this.vue, entity.status.error);
                }
                entity.updating = false;
            }

            if (!isSuccess(response)) {
                throw new Error(response.errors[0].detail);
            }
        }

        requestQueue:any[] = [];
        async callServerAuthed(path:string, args?:Object) : Promise<APISuccessResponse|APIErrorResponse> {
            return new Promise<APISuccessResponse|APIErrorResponse>(async (resolve) => {
                this.requestQueue.push({path:path, args:args, resolve:resolve});
                if (this.requestQueue.length > 1) {
                    return;
                }

                while (this.requestQueue.length > 0) {
                    let item = this.requestQueue[0];
                    let response = await callServer(item.path, item.args, this.authToken);
                    if (!isSuccess(response) && response.errors[0].status == 401 && this.onautherror) {
                        this.authToken = await new Promise<string>((resolve, reject) => {
                            this.onautherror(resolve);
                        });
                        // only 1 retry so we don't get stuck in infinite loop.
                        response = await callServer(item.path, item.args, this.authToken);
                    }
                    item.resolve(response);
                    this.requestQueue.shift();
                }
            });
        }
    }

    class Entity {
        ref:string;
        parent:Namespace;
        URLName:string;
        spec:EntitySpec;
        options:EntityOptions;
        args:EntityArgs;
        status:EntityStatus;
        loadRequested:boolean = false;
        loadRequestSequence = 0;
        autosaveSequence = 0;
        keySequence = 0;
        rows:Row[];
        singleRow:Row;
        removed:Editable[] = [];
        updating:boolean = false;
        revertJSON:string = '';

        onload?:Function;
        onloaderror?:Function;
        onsave?:Function;
        onsaveerror?:Function;

        constructor(ref:EntityRef, ns:Namespace, spec:EntitySpec, options:EntityOptions) {
            this.ref = ref;
            this.parent = ns;
            this.URLName = options.entity!;
            if (options.autosave) options.editable = true;
            if (options.editable) {
                if (spec.type != 'table') {
                    throw new Error(`Entity ${this.URLName} cannot be editable because it is type '${spec.type}'`);
                }
                if (!spec.identity.length) {
                    throw new Error(`Entity ${this.URLName} cannot be editable because it has no primary key`);
                }
            }

            this.spec = spec;
            this.options = options;
            this.args = options.args || {};
            this.status = {
                unused: true,
                edited: false,
                loading: false,
                saving: false,
                complete: false,
                error: "",
                pageCount: 1,
            }
            this.onload = options.onload;
            this.onloaderror = options.onloaderror;
            this.onsave = options.onsave;
            this.onsaveerror = options.onsaveerror;
            
            let params = this.parameterSpecs();
            for (var i = 0; i < params.length; i++) {
                var pname = params[i].spec.name;
                // make our properties reactive
                if (this.args[pname] === undefined) {
                    this.args[pname] = undefined;
                }
            }

            this.updating = true;
            if (this.options.single) {
                this.singleRow = this.initRow({});
                this.singleRow.$new = true;
            }
            this.rows = this.initRows([]);
            this.updating = false;
        }

        configureResult() {
            this.extendRows(this.rows);
            
            // for a single result, we don't want to track changes to a row object.
            // this prevents our single row from being marked as non-alive.
            if (this.options.editable && !this.options.single) {
                this.observeRows(<Editable[]>this.rows);
            }
        }

        loadOp() : string {
            return this.spec.operations.select ? "select" : "call";
        }

        parameterSpecs() : ParameterSpec[] {
            return this.spec.operations[this.loadOp()].parameters;
        }

        makeLoadBatchItems(args:EntityArgs) : [BatchEntry] {
            let outArgs:any[] = [];
            let params = this.parameterSpecs();
            for (var i = 0; i < params.length; i++) {
                var val = args[params[i].spec.name];
                outArgs.push(val);
            }
            let ret:[BatchEntry] = [[this.URLName + "." + this.loadOp(), outArgs]];
            if (this.loadOp() == "select" && args.options && args.options.pageSize) {
                // we're paging results, we need a count to figure out pageCount
                ret.push([this.URLName + ".count", [args.filter]]);
            }
            return ret;
        }

        processLoadResponseItems(items:any[]) : Row[]|Row {
            if (items.length > 1) {
                // currently this only happens when we're paging results - see makeLoadBatchItems above
                // remember, our count result hasn't been objectified...
                this.status.pageCount = Math.ceil(items[1].rows[0][0] / this.args.options.pageSize);
            }
            // modify our vue data object in place so we don't break variable references to it...
            return this.replaceResult(objectifyRows(items[0], () => this.initRow({})));
        }

        makeTruncateBatchItem() : BatchEntry {
            return [this.URLName + ".truncate", []];
        }

        persistedRowIdentity(row:Row) : any[] {
            let identity:any[] = [];
            for (let ispec of this.spec.identity) {
                if (row.$previous && typeof row.$previous[ispec.name] != 'undefined') {
                    identity.push(row.$previous[ispec.name]);
                } else {
                    identity.push(row[ispec.name]);
                }
            }
            return identity;
        }

        editedRowIdentity(row:Row) : any[] {
            let identity:any[] = [];
            for (let ispec of this.spec.identity) {
                identity.push(row[ispec.name]);
            }
            return identity;
        }

        editedRowIdentityCount(checkRow:Row) : number {
            let testStr = this.editedRowIdentity(checkRow).toString();
            let count = 0;
            for (let testRow of this.rows) {
                if (this.editedRowIdentity(testRow).toString() == testStr) {
                    count++;
                }
            }
            return count;
        }

        makeInsertBatchItem(row:Editable) : BatchEntry {
            if (!row.$new) {
                throw new Error("Transit internal error: attempting to insert item that is already persisted");
            }
            let identity = this.editedRowIdentity(row);
            let args = {};
            for (var n in row) {
                args[n] = row[n];
            }
            for (let ispec of this.spec.identity) {
                delete args[ispec.name];
            }
            return [`${this.URLName}.insert`, (<any[]>identity).concat([args])];
        }

        makeUpdateBatchItem(row:Editable) : BatchEntry {
            if (!row.$previous) {
                throw new Error("Transit internal error: attempting to save item that hasn't changed");
            }
            let identity = this.persistedRowIdentity(row);
            let args = {};
            // only grab the attributes that have changed,
            // but grab the new values...
            for (var n in row.$previous!) {
                args[n] = row[n];
            }
            return [`${this.URLName}.update`, (<any[]>identity).concat([args])];
        }

        makeDeleteBatchItem(row:Editable) : BatchEntry {
            return [`${this.URLName}.delete`, this.persistedRowIdentity(row)];
        }

        result() : Row[]|Row {
            return this.options.single ? this.singleRow : this.rows;
        }

        resultAsRows() : Row[] {
            return this.rows;
        }

        markRevertPoint() { 
            this.removed = [];
            this.revertJSON = JSON.stringify(this.rows);
        }

        replaceResult(rows:any[]) : Row[]|Row {
			this.rows["replace"](rows);
            if (rows["columns"]) this.rows["columns"] = rows["columns"]; // check for columns existence in case of truncate
            if (this.options.single) {
                for (let n in this.singleRow) {
                    this.singleRow[n] = undefined;
                }
                if (this.rows.length) {
                    for (let n in rows[0]) {
                        this.singleRow[n] = rows[0][n];
                    }
                    this.rows[0] = this.singleRow;
                    this.singleRow.$new = false;
                } else {
                    this.singleRow.$new = true;
                }
                this.singleRow.$previous = null;
            }
            this.markRevertPoint();
            this.removed = [];
            this.status.edited = false;
            return this.result();
        }

        async markEdited() {
            if (this.updating) return;
            this.status.edited = true;
            if (this.options.single && this.rows.length == 0) {
                this.rows.push(this.singleRow);
            }
            if (this.options.autosave) {
                let seq = ++this.autosaveSequence;
                do {
                    await delay(1);
                    if (this.autosaveSequence != seq) {
                        // a later autosave was requested - give up.
                        return;
                    }
                } while (this.status.saving);
                await this.parent.save(this.ref).catch(() => {});
            }
        }

        async revert() {
            this.updating = true;
            if (this.revertJSON == "") {
                this.replaceResult([]);
            } else {
                let rows = JSON.parse(this.revertJSON);
                this.replaceResult(rows);
            }
            this.updating = false;
            this.status.edited = false;
        }

        initRow(row:{[key:string]:any}) : {[key:string]:any} {
            // Our properties get defined *before* Vue defines its properties.
            // Don't redefine our props (say upon row re-insertion), because Vue 
            // won't redefine props either and our re-define will destroy Vue's reactivity.
            if (!row.$ref) {
                for (let output of this.spec.operations[this.loadOp()].outputs) {
                    if (this.options.editable) {
                        trackChanges(row, output.spec.name, row[output.spec.name], this);
                    } else {
                        Object.defineProperty(row, output.spec.name, { configurable: true, writable: true, enumerable: true, value: row[output.spec.name] });
                    }
                }
            }
            if (this.options.editable) {
                Object.defineProperty(row, "$ref", { configurable: true, writable: true, enumerable: false, value: this.ref });
                Object.defineProperty(row, "$previous", { configurable: true, writable: true, enumerable: false, value: null });
                Object.defineProperty(row, "$new", { configurable: true, writable: true, enumerable: false, value: false });
                Object.defineProperty(row, "$alive", { configurable: true, writable: true, enumerable: false, value: true });
                Object.defineProperty(row, "$key", { configurable: true, writable: true, enumerable: false, value: this.keySequence++ });
            }
            return row;
        }

        initRows(rows:Row[]) : Row[] {    
            Object.defineProperty(rows, "$ref", { configurable: true, writable: true, enumerable: false, value: this.ref });
            return rows;
        }

		extendRows(rows:Row[]) {
            // This is a workaround for JavaScript engine limitations on
            // the number of parameters to a function. For an extremely large
            // result set, the rows are splatted as arguments to splice, which
            // triggers an error in most engines. By using this method instead
            // we circumvent the need to use splice to replace an array's contents.
            Object.defineProperty(
            	rows,
            	"replace",
            	{
	            	value: function(newRows) {
                        rows.splice(0); // this will trigger Vue reactivity if we currently have rows,
                                        // for the case where newRows is empty so we still trigger
						for(var i = 0; i < newRows.length; i++) {
                            window["Vue"].set(rows, i, newRows[i]); // this will trigger Vue reactivity
						}
            		},
            		configurable: true,
            		writable: true,
            		enumerable: false
            	}
            )
		}
		
        observeRows(rows:Editable[]) {
            let markRemoved = (row:Editable) => {
                row.$alive = false;
                if (!this.updating && !row.$new) {
                    this.markEdited();
                    this.removed.push(row);
                }
            }

            var orig = {
                push: rows.push,
                pop: rows.pop,
                shift: rows.shift,
                unshift: rows.unshift,
                splice: rows.splice,
                replace: rows["replace"],
            };
    
            let addRows = (rowSet, start?) => {
                start = start || 0;
                for (var i = start; i < rowSet.length; i++) {
                    if (typeof rowSet[i] != "object" || rowSet[i] == null) {
                        throw new Error("transit row element must be an object");
                    }
                    this.initRow(rowSet[i]);
                    if (!this.updating) {
                        this.markEdited();
                        rowSet[i].$new = true;
                    }
                }
                return rowSet;
            }
    
            rows.push = function() {
                addRows(arguments);
                return orig.push.apply(this, arguments);
            }
            rows.pop = function() {
                if (rows.length) {
                    markRemoved(rows[rows.length-1]);
                }
                return orig.pop.apply(this, arguments);
            }
            rows.unshift = function() {
                addRows(arguments);
                return orig.unshift.apply(this, arguments);
            }
            rows.shift = function() {
                if (rows.length) {
                    markRemoved(rows[0]);
                }
                return orig.shift.apply(this, arguments);
            }
            rows.splice = function(start, length) {
                start = start || 0;
                length = length || 0;
                var removed = rows.slice(start, start + length);
                for (var i = 0; i < removed.length; i++) {
                    markRemoved(removed[i]);
                }
                addRows(arguments, 2);
                return orig.splice.apply(this, arguments);
            }
			rows["replace"] = function(newRows) {
				for (var i = 0; i < rows.length; i++) {
					markRemoved(rows[i]);
				}
				addRows(newRows);
				return orig.replace.call(this, newRows);
			}
    
            return rows;
        }

        csv() {
            function escapeCSV(value) {
                value = (value == null || value == undefined) ? "" : String(value);
                if (value.match(/[,\n]/)) {
                    value = value.replace(/"/g, '""');
                    value = '"' + value + '"';
                }
                return value;
            }
            
            var outrows:any[] = [];
            var row:any = [];
            var cols = this.rows["columns"];
            for (var i = 0; i < cols.length; i++) {
                row.push(escapeCSV(cols[i].spec.name));
            }
            outrows.push(row.join());
            for (var i = 0; i < this.rows.length; i++) {
                row = [];
                for (var j = 0; j < cols.length; j++) {
                    row.push(escapeCSV(this.rows[i][cols[j].spec.name]));
                }
                outrows.push(row.join());
            }
            outrows.push(""); // adds trailing newline
            return outrows.join("\n");
        }
    }

    export function callServer(path:string, args?:Object, authToken?:AuthToken) : Promise<APISuccessResponse|APIErrorResponse> {
        return new Promise(function(resolve, reject) {
            var oXHR = new XMLHttpRequest();
            oXHR.open("POST", "/v1/" + path, true);
            oXHR.setRequestHeader("Content-Type", "application/json;charset=UTF-8");
            if (authToken) {
                oXHR.setRequestHeader("x-transit-authorization", authToken);
            }
            oXHR.onreadystatechange = function (oEvent) { 
                if (this.readyState === 4) {
                    let json;
                    try {
                        json = JSON.parse(this.responseText);
                    } catch(e) {
                        json = { errors: [{ status: -1, detail: this.responseText || "Unexpected server response" }]};
                    }
                    resolve(json);
                }  
            }; 
            oXHR.send(JSON.stringify(args||[]));
        });
    };

    export function isSuccess(response:APISuccessResponse|APIErrorResponse) : response is APISuccessResponse {
        return typeof (<APISuccessResponse>response).data != 'undefined';
    }

    export function delay(ms:number) {
        return new Promise((resolve) => setTimeout(resolve, ms));
    }

    function objectifyRows(result:DatabaseResult, rowConstructor:() => Row) : Row[] {
        // turn the row arrays into objects/hashes
        let outRows : any[] = [];
        for (let inRow of result.rows) {
            var outRow = rowConstructor();
            for (var j = 0; j < result.columns.length; j++) {
                outRow[result.columns[j].spec.name] = inRow[j];
            }
            outRows.push(outRow);
        }
        outRows["columns"] = result.columns;
        return outRows;
    }

    function trackChanges(obj:Object, key:string, val:any, ent:Entity) {
        Object.defineProperty(obj, key, {
            configurable: true,
            enumerable: true,
            get: function() {
                return val;
            },
            set: function(newVal) {
                if (newVal === val) return; 
                if (this.$alive && !ent.updating) {
                    ent.markEdited();
                }
                if (!this.$previous) {
                    this.$previous = {};
                }
                this.$previous[key] = val;
                val = newVal;
            }
        });
    }
}

/**API_CONSTRUCTOR**/
