/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only. 
 */

import * as http from "http";
import * as server from '../lib/Server';
import { DataType } from "../lib/Command";
import * as log from '../lib/Logger';
import * as fs from "fs";
import "jest";
import { SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION } from "constants";

const PORT = 8001;
const sqlServerConfigPath = "src/test/Server/config.json";

var XMLHttpRequest : any;

function asserteq(a, b) {
    expect(a).toBe(b);
}

async function startServer(config) : Promise<server.Server> {
    (log as any).enabled = false;
    let serv = await server.startServer(config);
    (log as any).enabled = true;
    return serv;
}

test("retrieve namespaces", async function() {
    let serv = await startServer(sqlServerConfigPath);
    try {
        let response = await callServer("/v1/namespaces", {}, [], 200, "application/json");
        expect(JSON.parse(response).data).toEqual(["sqlserver", "decl"]);
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});

test("SQL Server view", async function() {
    let serv = await startServer(sqlServerConfigPath);
    try {
        let data = await callServer("/v1/sqlserver/All_Datatypes_View/select", {}, [], 200, "application/json");
        checkResponse(data);
        data = await callServer("/v1/sqlserver/All_Datatypes_View/select", {}, [["col bigint", "col varchar"]], 200, "application/json");
        checkResponse(data);
        expect(JSON.parse(data).data.columns.map((col) => col.spec.name)).toEqual(["col bigint", "col varchar"]);
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});

test("SQL Server procedure", async function() {
    let serv = await startServer(sqlServerConfigPath);
    try {
        let data = await callServer('/v1/sqlserver/echo_types/call', {}, [
            3, 
            true, 
            'xyz', 
            '2018-01-01', 
            '2018-01-01T13:12:11.000', 
            '2018-01-01T13:12:11.0000000', 
            '2018-01-01T08:12:11.0000000-05:00',
            3.14,
            3.14,
            3,
            3.14,
            'xyz',
            'xyz',
            3.14,
            'xyz',
            3.14,
            '2018-01-01T13:12:00',
            3,
            3.14,
            'xyz',
            '13:12:11',
            3,
            'c15c77f2-aaeb-4629-84cf-76aad3746c8c',
            'xyz'
        ], 200, "application/json");
        checkResponse(data);
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});

test("SQL Server function (fn)", async function() {
    let runtime = await server.startRuntime(sqlServerConfigPath);
    let api = runtime.getAPI("sqlserver");
    asserteq(api.$name, "sqlserver");
    let result = await api.echo_fn.call("foo");
    asserteq(result[0].result, "foo");
});

test("SQL Server table function (tf)", async function() {
    let runtime = await server.startRuntime(sqlServerConfigPath);
    let api = runtime.getAPI("sqlserver");
    let result = await api.echo_tf.call("foo", 3);
    asserteq(result[0].Msg, "foo");
    asserteq(result[0].Num, 3);
});

test("web serving", async function() {
    let serv = await startServer(sqlServerConfigPath);
    try {
        let data2 = await callServer("/v1/sqlserver/All_Datatypes_View/select", {}, [], 200, "application/json");
        let data = await callServer('/test.txt', {}, null, 200, "text/plain");
        asserteq(data, "static content");
        data = await callServer("/", {}, null, 200, "text/html");
        asserteq(data, "index");

        data = await callServer('/v1/sqlserver.html', {}, null, 200, "text/html");
        expect(data).toMatch(/sqlserver\.js/);

        // test caching
        let result = await newServerPromise('/test.txt', {}, null);
        let mod = Date.parse(result.response.headers['last-modified']);
        expect(mod).not.toBe(NaN);
        
        data = await callServer('/test.txt', { 'If-Modified-Since': new Date(mod).toUTCString() }, null, 304, undefined);
        asserteq(data, "");

        data = await callServer('/test.txt', { 'If-Modified-Since': new Date(mod - 1000).toUTCString() }, null, 200, "text/plain");
        asserteq(data, "static content");

        data = await callServer('/test.txt', { 'If-Modified-Since': new Date(mod + 1000).toUTCString() }, null, 304, undefined);
        asserteq(data, "");

        result = await newServerPromise('/include_test.html', {}, null);
        let incmod = Date.parse(result.response.headers['last-modified']);
        asserteq(result.data, "fooincludebarincludebaz");

        await new Promise((resolve, reject) => {
            setTimeout(resolve, 2000);
        });

        // update the modified time of the include file, make sure last-modified changes
        fs.futimesSync(fs.openSync('src/test/Server/include_test.inc.html', 'a'), new Date(), new Date());

        result = await newServerPromise('/include_test.html', {}, null);
        let incmod2 = Date.parse(result.response.headers['last-modified']);
        expect(incmod2).toBeGreaterThan(incmod);
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});


test("custom driver import", async function() {
    // test hard-coded path to plugin database driver.
    let config : server.ServerConfig = JSON.parse(fs.readFileSync(sqlServerConfigPath).toString());
    config.namespaces[0].driver = "../lib/driver/SQLServer";
    let serv = await startServer(config);
    asserteq(true, true); // server is running
    await new Promise((resolve) => serv.close(resolve));
});

test("api javascript", async function() {
    let serv = await startServer(sqlServerConfigPath);
    try {
        let data = await callServer("/v1/$clientConfig.js", {}, null, 200, "text/plain");
        var transitClientConfig : any = null;
        eval(data);
        expect(transitClientConfig).not.toBeNull();

        data = await callServer("/v1/sqlserver.js", {}, null, 200, "text/plain");
        var sqlserver : any = null;
        eval(data);
        expect(sqlserver).not.toBeNull();

        await checkAPI(sqlserver);
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});


test("standalone runtime", async function() {
    let runtime = await server.startRuntime(sqlServerConfigPath);
    let result = await runtime.executeSQL("sqlserver", "SELECT 5 AS five");
    asserteq(result.length, 1);
    asserteq(result[0].five, 5);

    result = await runtime.executeSQL("sqlserver", "SELECT ? as one, ? as two, '??' as three, ?", "foo'bar", "baz''bee", 5);
    asserteq(result[0].one, "foo'bar");
    asserteq(result[0].two, "baz''bee");
    asserteq(result[0].three, "?");
    asserteq(result[0][""], 5);

    let api = runtime.getAPI("sqlserver");
    checkAPI(api);
});

test("table CRUD", async function() {
    let runtime = await server.startRuntime(sqlServerConfigPath);
    function checkRow(actual, expected) {
        asserteq(actual.pk1, expected.pk1);
        asserteq(actual.pk2, expected.pk2);
        asserteq(actual.int_null, expected.int_null);
        asserteq(actual.int_not_null_default_0, expected.int_not_null_default_0);
        asserteq(actual.int_not_null, expected.int_not_null);
    }

    let api = runtime.getAPI("sqlserver");
    await api.test_table.truncate();

    // note that column 'str' is omitted...
    let result = await api.test_table.insert(1, 2, { int_null: 3, int_not_null_default_0: 4, int_not_null: 5 });
    asserteq(result.length, 0);

    result = await api.test_table.select();
    asserteq(result.length, 1);
    checkRow(result[0], { pk1: 1, pk2: 2, int_null: 3, int_not_null_default_0: 4, int_not_null: 5 });

    result = await api.test_table.update(1, 2, { pk1: 6, pk2: 7, int_null: 8, int_not_null_default_0: 9, int_not_null: 10 });
    asserteq(result.length, 0);

    // note that str is null here - legit operation...
    result = await api.test_table.update(1, 2, { pk1: 6, pk2: 7, int_null: 7, int_not_null_default_0: 9, int_not_null: 10 });
    asserteq(result.length, 0);

    // we shouldn't have to supply a "required" arg to an update...
    result = await api.test_table.update(6, 7, { int_null: 8, int_not_null_default_0: 9 });
    asserteq(result.length, 0);

    result = await api.test_table.select();
    checkRow(result[0], { pk1: 6, pk2: 7, int_null: 8, int_not_null_default_0: 9, int_not_null: 10 });

    await api.test_table.insert(1, 2, { int_null: 3, int_not_null_default_0: 4, int_not_null: 5 });
    result = await api.test_table.select(['int_null'], { pk1: 1, pk2: 2 }, ["pk1", "pk2"]);
    checkRow(result[0], { int_null: 3 });
    asserteq(result.length, 1);

    result = await api.test_table.select(['int_null'], [ ["pk1", "=", 1], ["pk2", "=", 2] ], ["pk1", "pk2"]);
    checkRow(result[0], { int_null: 3 });
    asserteq(result.length, 1);

    result = await api.test_table.delete(6, 7);
    asserteq(result.length, 0);
    
    result = await api.test_table.select();
    asserteq(result.length, 1);

    await api.test_table.truncate();
    result = await api.test_table.select();
    asserteq(result.length, 0);    
});

test("Identity columns", async () => {
    let runtime = await server.startRuntime(sqlServerConfigPath);
    let api = runtime.getAPI("sqlserver");

    await api.ident.truncate();
    let result = await api.ident.insert(null, { str: "foo" });

    // TODO: server returns row with identity in it

    result = await api.ident.update(1, { str: "foo2" });

    result = await api.ident.select();
    expect(result.length).toBe(1);
    expect(result[0].str).toBe("foo2");
});

test("Sorting", async function() {
    let runtime = await server.startRuntime(sqlServerConfigPath);
    await runtime.executeSQL("sqlserver", "DELETE FROM simple;");

    let api = runtime.getAPI("sqlserver");
    await api.simple.insert(1, 3, { str: "row 1" });
    await api.simple.insert(1, 1, { str: "row 1" });
    await api.simple.insert(1, 2, { str: "row 1" });

    let result = await api.simple.select(null, null, ["pk2"]);
    expect(result[0].pk2).toBe(1);
    expect(result[1].pk2).toBe(2);
    expect(result[2].pk2).toBe(3);

    result = await api.simple.select(null, null, ["!pk2"]);
    expect(result[0].pk2).toBe(3);
    expect(result[1].pk2).toBe(2);
    expect(result[2].pk2).toBe(1);
});

test("In operator", async function() {
    let runtime = await server.startRuntime(sqlServerConfigPath);
    let api = runtime.getAPI("sqlserver");
    await api.simple.truncate();

    await api.simple.insert(1, 1, { str: "row 1" });
    await api.simple.insert(1, 2, { str: "row 2" });
    await api.simple.insert(1, 3, { str: "row 3" });

    let result = await api.simple.select(null, [ ["pk2", "in", [2, 3]] ], ["pk2"]);
    expect(result.length).toBe(2);
    expect(result[0].pk2).toBe(2);
    expect(result[1].pk2).toBe(3);

    result = await api.simple.select(null, {pk2: [1, 3]});
    expect(result.length).toBe(2);
    expect(result[0].pk2).toBe(1);
    expect(result[1].pk2).toBe(3);

    result = await api.simple.select(null, {pk2: []});
    expect(result.length).toBe(0);
});

test("graceful failures", async function() {
    let api = await server.generateAPI(sqlServerConfigPath, "sqlserver");
    let errorMsg = '';
    // this will generate an error, let's make sure the error is appropriately descriptive...
    try {
        let result = await api.test_table.insert(1, 2, { int_null: 3, int_not_null_default_0: 4, int_not_null: 5 , foo: "bar" });
    } catch(e) {
        errorMsg = e.toString();
    }
    expect(errorMsg).toContain(`Unknown object key "foo"`);
});

test("identity", async function() {
    // test hard-coded path to plugin database driver.
    let config : server.ServerConfig = JSON.parse(fs.readFileSync(sqlServerConfigPath).toString());
    config.identity = {
        driver: "Simple",
        config: {
            users: [
                {
                    username: "me",
                    password: "mypass",
                    roles: ["user", "friend"]
                },
                {
                    username: "2ndme",
                    password: "mypass",
                    roles: ["vagrant"]
                }
            ]
        }
    };
    config.namespaces[0].allowedUsernames = ["*"];
    let serv = await startServer(config);

    function newCreds(username, password) : string {
        let creds = Buffer.from(JSON.stringify({username: username, password: password})).toString('base64');
        return Buffer.from(JSON.stringify({username: username, password: password})).toString('base64');
    }

    try {
        await callServer("/v1/sqlserver/some_entity", {}, [], 401, "application/json");
        await callServer("/v1/sqlserver/some_entity", { "x-transit-authorization": newCreds("notme", "mypass") }, [], 401, "application/json");
        await callServer("/v1/sqlserver/some_entity", { "x-transit-authorization": newCreds("me", "notmypass") }, [], 401, "application/json");
        // 404 is a success, the entity just doesn't exist
        await callServer("/v1/sqlserver/some_entity", { "x-transit-authorization": newCreds("me", "mypass") }, [], 404, "application/json");
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }

    config.namespaces[0].allowedUsernames = ["me"];
    serv = await startServer(config);
    try {
        await callServer("/v1/sqlserver/some_entity", { "x-transit-authorization": newCreds("2ndme", "mypass") }, [], 401, "application/json");
        await callServer("/v1/sqlserver/some_entity", { "x-transit-authorization": newCreds("me", "mypass") }, [], 404, "application/json");
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }

    config.namespaces[0].allowedUsernames = [];
    config.namespaces[0].allowedRoles = ["nobody"];
    serv = await startServer(config);
    try {
        await callServer("/v1/sqlserver/some_entity", { "x-transit-authorization": newCreds("me", "mypass") }, [], 401, "application/json");
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }

    config.namespaces[0].allowedRoles = ["friend"];
    serv = await startServer(config);
    try {
        await callServer("/v1/sqlserver/some_entity", { "x-transit-authorization": newCreds("me", "mypass") }, [], 404, "application/json");
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});

test("count API", async function() {
    let runtime = await server.startRuntime(sqlServerConfigPath);
    let api = runtime.getAPI("sqlserver");
    let result = await api.many_rows.count();
    expect(result[0].count).toBe(70000);
    result = await api.many_rows.count([["num", ">", 200],["num", "<", 300]]);
    expect(result[0].count).toBe(99);
});

test("result paging", async function() {
    let runtime = await server.startRuntime(sqlServerConfigPath);
    let api = runtime.getAPI("sqlserver");

    let result = await api.many_rows.select(null, null, ["num"], {pageSize:3,pageIndex:1});
    expect(result.length).toBe(3);
    expect(result[0].num).toBe(0);
    expect(result[1].num).toBe(1);
    expect(result[2].num).toBe(2);

    result = await api.many_rows.select(null, null, ["num"], {pageSize:3,pageIndex:3001});
    expect(result.length).toBe(3);
    expect(result[0].num).toBe(9000);
    expect(result[1].num).toBe(9001);
    expect(result[2].num).toBe(9002);

    result = await api.many_rows.select(null, null, ["num"], {pageSize:3,pageIndex:23334});
    expect(result.length).toBe(1);
    expect(result[0].num).toBe(69999);

    result = await api.many_rows.select(null, null, ["num"], {pageSize:3,pageIndex:23335});
    expect(result.length).toBe(0);

    let error = false;
    try {
        result = await api.many_rows.select(null, null, ["num"], {pageSize:3,pageIndex:-1});
    } catch(e) {
        error = true;
    }
    expect(error).toBe(true);
});

test("batch API", async function() {
    let serv = await startServer(sqlServerConfigPath);
    try {
        let data = await callServer("/v1/sqlserver/$batch-api", {}, [
            ["All_Datatypes_View.select", []],
            ["All_Datatypes_View.select", [["col bigint", "col varchar"]]],
        ], 200, "application/json");
        let arr = JSON.parse(data).data;
        checkResponse(arr[0]);
        checkResponse(arr[1]);
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});

test("batch transaction", async function() {
    let serv = await startServer(sqlServerConfigPath);
    try {
        let data = await callServer("/v1/sqlserver/$batch-api", {}, [
            ["All_Datatypes_View.select", []],
            ["All_Datatypes_View.select", [["col bigint", "col varchar"]]],
        ], 200, "application/json");
        let arr = JSON.parse(data).data;
        checkResponse(arr[0]);
        checkResponse(arr[1]);
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});

test("batch API javascript", async function() {
    let serv = await startServer(sqlServerConfigPath);
    try {
        let data = await callServer("/v1/sqlserver.js", {}, null, 200, "text/plain");
        var sqlserver : any = null;
        eval(data);
        expect(sqlserver).not.toBeNull();

        await sqlserver.simple.truncate();

        // affirm the positive case
        await sqlserver.$batch_api([
            ["simple.insert", [1, 1, {}]],
            ["simple.insert", [1, 2, {}]]
        ]);
        let result = await sqlserver.simple.select();
        expect(result.length).toBe(2);

        await sqlserver.simple.truncate();

        // make sure a failure rolls back
        // this will log an error, so turn off logging
        (log as any).enabled = false;
        try {
            await sqlserver.$batch_api([
                ["simple.insert", [1, 1, {}]],
                ["simple.insert", [1, 2, {}]],
                ["simple.insert", [1, 1, {}]] // will trigger error on primary key
            ]);
        } catch (e) {
            let result = await sqlserver.simple.select();
            expect(result.length).toBe(0);
        }
        (log as any).enabled = true;

        await sqlserver.simple.truncate();
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});

test("Table loading", async function() {
    let runtime = await server.startRuntime(sqlServerConfigPath);
    let api = runtime.getAPI("sqlserver");

    await api.simple.truncate();

    await api.simple.insertRowsFlat([
        ["pk1", "pk2", "str"],
        [1, 1, "row 1" ],
        [2, 2, "row 2" ]
    ]);
    let result = await api.simple.select(null, null, ["pk1"]);
    expect(result.length).toBe(2);
    expect(result[0].pk1).toBe(1);
    expect(result[1].pk1).toBe(2);

    await api.simple.insertRowsFlat([
        ["pk2", "pk1", "str", "num"],
        ["1", "3", "row 1", null],
        ["2", "4", "row 2", undefined]
    ]);
    result = await api.simple.select(null, null, ["pk1"]);
    expect(result.length).toBe(4);
    expect(result[0].pk1).toBe(1);
    expect(result[1].pk1).toBe(2);
    expect(result[2].pk1).toBe(3);
    expect(result[3].pk1).toBe(4);

    await api.simple.replaceRowsFlat([
        ["PK2", "Pk1", "StR", "num"],
        [1, 4, "row 1", null],
        [2, 5, "row 2", undefined]
    ]);
    result = await api.simple.select(null, null, ["pk1"]);
    expect(result.length).toBe(2);
    expect(result[0].pk1).toBe(4);
    expect(result[1].pk1).toBe(5);
});

test("Declared driver", async function() {
    let serv = await startServer(sqlServerConfigPath);
    try {
        let data = await callServer("/v1/decl.js", {}, null, 200, "text/plain");
        var decl : any = null;
        eval(data);
        expect(decl).not.toBeNull();

        let result = await decl.echo.call("foo");
        asserteq(result, "foo");

        result = await decl.echoNum.call(5);
        asserteq(result, 5);

        result = await decl.nothing.call();
        asserteq(result, null);

        result = await decl.$batch_api([
            ["echo.call", ["foo"]],
            ["echo.call", ["bar"]]
        ]);
        asserteq(result[0], "foo");
        asserteq(result[1], "bar");
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});

test("Extensions", async function() {
    let serv = await startServer(sqlServerConfigPath);
    try {
        let data = await callServer("/v1/sqlserver.js", {}, null, 200, "text/plain");
        var sqlserver : any = null;
        eval(data);

        let result = await sqlserver.echo.call("foo");
        asserteq(result, "foo");

        result = await sqlserver.callEcho.call("foo");
        asserteq(result, "foo");
    } finally {
        await new Promise((resolve) => serv.close(resolve));
    }
});

async function checkAPI(api: any) {
    let data = await api.All_Datatypes.select();
    checkResponse(data);

    data = await api.All_Datatypes_View.select();
    checkResponse(data);
}

function checkResponse(dataStr: any) {
    if (typeof dataStr == "object") {
        var data : any = dataStr;
    } else {
        var data : any = JSON.parse(dataStr).data;
    }

    // tedious will convert date values to Date objects, which means our standalone
    // runtime API will do the same, in contrast to the http API.
    function dateString(d) { return typeof d == "string" ? d : d.toISOString(); }
    for (let i = 0; i < data.columns.length; i++) {
        var col = data.columns[i];
        var val = data.rows ? data.rows[0][i] : data[0][col.spec.name];
        switch (col.spec.dataType) {
            case DataType.Boolean:
                expect(val).toBe(true);
                break;
            
            case DataType.Integer:
                if (col.spec.name == "col bigint") expect(val).toBe("3");
                else expect(val).toBe(3);
                break;
            
            case DataType.Float:
                if (col.spec.name == "col real") expect(val).toBe(3.140000104904175);
                else expect(val).toBe(3.14);
                break;
            
            case DataType.String:
                switch (col.spec.name) {
                    case "col bigint":
                        expect(val).toBe("123");
                        break;
                    
                    case "col uniqueidentifier":
                        expect(val.toLowerCase()).toBe("c15c77f2-aaeb-4629-84cf-76aad3746c8c");
                        break;
                    
                    default:
                        expect(val).toBe("xyz");
                        break;
                }
                break;

            case DataType.Date:
                expect(dateString(val)).toBe('2018-01-01T00:00:00.000Z');
                break;

            case DataType.DateTime:
                if (col.spec.name == "col smalldatetime") expect(dateString(val)).toBe('2018-01-01T13:12:00.000Z');
                else if (col.spec.name == "col time") expect(dateString(val)).toBe('1970-01-01T13:12:11.000Z');
                else expect(dateString(val)).toBe('2018-01-01T13:12:11.000Z');
                break;
            
            default:
                throw new Error(`Internal error: unknown datatype ${col.spec.dataType}`);
        }
    }
}

async function callServer(path : string, headers: { [key: string] : any }, postVars : null | { [key: string] : any }, expectedStatus : number, expectedContentType : string|undefined) : Promise<string> {
    let result = await newServerPromise(path, headers, postVars);
    expect(result.response.statusCode).toBe(expectedStatus);
    expect(result.response.headers['content-type']).toBe(expectedContentType);
    return result.data;
}

function newServerPromise(path : string, headers: { [key: string] : any }, vars : null | { [key: string] : any }) : Promise<any> {
    return new Promise((resolve, reject) => {

        let method = (vars === null ? "GET" : "POST");
        let postBody = JSON.stringify(vars);
        if (method == "POST") {
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded',
                'Content-Length': Buffer.byteLength(postBody),
                ...headers
            };
        }

        var req = http.request(
            { 
                host: "localhost", 
                port: PORT, 
                path: path, 
                method: method, 
                headers: headers
            }, 
            function (res : http.IncomingMessage) {
                let data = '';
                        
                res.setEncoding('utf8');
                res.on('data', (chunk) => { data += chunk; });
                res.on('end', () => {
                    resolve({ response: res, data: data});
                });
            }
        );
            
        req.on("error", (err: Error) => {
            reject(err);
        });

        if (method == "POST") {
            req.write(postBody);
        }
        
        req.end();
    });
}

XMLHttpRequest = function() {
    XMLHttpRequest.promise = new Promise((resolve, reject) => {
        XMLHttpRequest.resolve = resolve;
    });
};
XMLHttpRequest.prototype = {
    headers: {},
    open: function(method, url, async, username, password) {
        var _this = this;
        this.send = function(body) {
            _this.responseText = "";
            var req = http.request(
                { 
                    host: "localhost", 
                    port: PORT, 
                    path: url, 
                    method: method, 
                    headers: this.headers
                }, 
                (res : http.IncomingMessage) => {
                    let data = '';
                    _this.status = res.statusCode;
                    _this.statusText = res.statusMessage;
                    res.setEncoding('utf8');
                    res.on('data', (chunk) => { _this.responseText += chunk; });
                    res.on('end', () => {
                        _this.readyState = 4;
                        _this.onreadystatechange();
                        XMLHttpRequest.resolve();
                    });
                }
            );
            
            req.on("error", (err: Error) => {
                _this.status = 500;
                _this.statusText = err.toString();
                _this.responseText = _this.statusText;
                _this.readyState = 4;
                _this.onreadstatechange();
                XMLHttpRequest.resolve();
            })

            req.write(body);
            req.end();
        }
    },
    setRequestHeader: function(header, value) {
        this.headers[header] = value;
    }
}
