import * as server from '../lib/Server';
import { JSDOM, VirtualConsole } from "jsdom";
import * as log from '../lib/Logger';

const PORT = 8002;
const configPath = "src/test/Vue/config.json";

var serv;
var window;
beforeEach(async function() {
    // inhibit the server-started log message...
    let log = console.log;
    console.log = () => {};
    serv = await server.startServer(configPath);
    console.log = log;
    serv.keepAliveTimeout = 250;

    //inhibit the vue dev mode log message...
    let info = console.info;
    console.info = () => {};
    window = await loadPage(`http://localhost:${PORT}/`);
    await sleep(1); // the vue log message triggers asynchronously after load, we need to wait one little bit longer....
    console.info = info;
});
afterEach(async function() {
    window.close();
    await new Promise((resolve, reject) => serv.close(resolve));
});

test("Vue db proc", async () => {
    let success = 0;
    window.data.echo.onExecute = function() {
        success++;
    }
    checkStatus(window.data.echo, true, false, false, false);
    window.data.echo.args.message = "test message";
    await sleep(1);
    checkStatus(window.data.echo, false, true, false, false);
    await sleep(100);
    expect(window.data.echo.rows[0].message).toBe("test message");
    expect(success).toBe(1);
    checkStatus(window.data.echo, false, false, true, false);
});

test("Vue function args", async () => {
    window.data.echo.args.message = function() { return "functional message"; }
    await sleep(1000);
    expect(window.data.echo.rows[0].message).toBe("functional message");
});

test("Vue operation availability", async () => {
    expect(window.data.All_Datatypes.select).not.toBeNull();
    expect(window.data.All_Datatypes.insert).not.toBeNull();
    expect(window.data.All_Datatypes.update).not.toBeNull();
    expect(window.data.All_Datatypes.delete).not.toBeNull();
});

test("Vue property naming", async () => {
    expect(window.data.foobar).not.toBeNull();
});

test("Vue db view", async () => {
    let entity = window.data.All_Datatypes;
    let resolved = false;
    checkStatus(entity, true, false, false, false);
    entity.execute().then(() => resolved = true);
    await sleep(1);
    checkStatus(entity, false, true, false, false);
    await sleep(100);
    expect(entity.rows[0]["col bigint"]).toBe("3");
    expect(entity.rows[0]["col varchar"]).toBe("xyz");
    expect(resolved).toBe(true);
    checkStatus(entity, false, false, true, false);
});

test("Vue single row result", async () => {
    let entity = window.data.All_Datatypes;
    entity.execute();
    await sleep(1000);
    expect(entity.row["col bigint"]).toBe("3");
    expect(entity.row["col varchar"]).toBe("xyz");
    entity.args.filter = { "col bigint": 5 } // will return 0 results
    await sleep(1000);
    expect(entity.row["col bigint"]).toBe(undefined);
});

test("Vue RPC failure", async() => {
    // turn off error logging because we're gonna generate some
    (log as any).enabled = false;

    let error = 0;
    window.data.echo.onError = function() {
        error++;
    }

    checkStatus(window.data.echo, true, false, false, false);
    window.data.echo.args.message = 1; // mismatched datatype which should trigger a server 500 error
    await sleep(1);
    checkStatus(window.data.echo, false, true, false, false);
    await sleep(100);
    checkStatus(window.data.echo, false, false, true, true);

    expect(error).toBe(1);

    (log as any).enabled = true;
});

test("Vue multiple calls", async() => {
    let origCallAPI = window.sqlserver._.callAPI;
    let callCount = 0;
    let updateCount = 0;
    window.sqlserver._.callAPI = function() {
        callCount++;
        return origCallAPI.apply(this, arguments);
    }
    window.data.echo.onExecute = function() {
        updateCount++;
    }

    // trigger a request
    window.data.echo.args.message = "test message 1";

    // trigger a request before the previous one has been sent
    window.data.echo.args.message = "test message 2";
    await sleep(1);

    // trigger a request after the previous one has been sent
    window.data.echo.args.message = "test message 3";

    await sleep(100);
    expect(callCount).toBe(2);
    expect(updateCount).toBe(1);
    expect(window.data.echo.rows[0].message).toBe("test message 3");
});

test("Vue table editing", async () => {
    async function checkResult(actual) {
        let rows = await window.sqlserver.simple.select(null, null, ["pk1","pk2"]);            
        expect(rows.length).toBe(actual.length);
        for (var i = 0; i < rows.length; i++) {
            expect(rows[i]).toEqual(actual[i].$transit._copy());
        }
    }
    
    function checkBooleans(row, isNew, editing, saving) {
        expect(row.$transit.isNew).toBe(isNew);
        expect(row.$transit.editing).toBe(editing);
        expect(row.$transit.saving).toBe(saving);
    }

    let tbl = window.app.simple;
    let api = window.sqlserver;
    await tbl.execute();
    while (tbl.rows.length) {
        await tbl.rows[0].$transit.removeAndSave();
    }
    await checkResult([]);

    let row1 : any = { pk1: 10, pk2: 1, str: "row 1", num: null };
    let row2 : any = { pk1: 10, pk2: 2, str: "row 2", num: 1 };
    let row3 : any = { pk1: 10, pk2: 3, str: "row 3", num: 2 };

    // test insert, push
    tbl.rows.push(row1);
    checkBooleans(row1, true, false, false);
    await tbl.rows[0].$transit.save();
    await checkResult([row1]);

    // test update
    tbl.rows[0].$transit.edit();
    checkBooleans(row1, false, true, false);
    row1.str = "row 1 edited";
    let prms = tbl.rows[0].$transit.save();
    checkBooleans(row1, false, false, true);
    await prms;
    await checkResult([row1]);
    checkBooleans(row1, false, false, false);
    
    // test cancel
    tbl.rows[0].$transit.edit();
    checkBooleans(row1, false, true, false);
    row1.str = "row 1";
    tbl.rows[0].$transit.cancel();
    expect(row1.str).toBe("row 1 edited");
    await checkResult([row1]);
    checkBooleans(row1, false, false, false);

    // test delete
    await tbl.rows[0].$transit.removeAndSave();

    // test splice
    tbl.rows.splice(0,0,row1,row2,row3);
    await tbl.rows[0].$transit.save();
    await checkResult([row1]);
    await tbl.rows[1].$transit.save();
    await checkResult([row1, row2]);
    await tbl.rows[2].$transit.save();
    await checkResult([row1, row2, row3]);

    tbl.rows.splice(1,1);
    await api.vueSave(tbl);
    await checkResult([row1, row3]);

    // test pop
    tbl.rows.pop();
    await api.vueSave(tbl);
    await checkResult([row1]);

    // test shift
    tbl.rows.shift();
    await api.vueSave(tbl);
    await checkResult([]);

    // test unshift
    tbl.rows.unshift(row1, row2, row3);
    await tbl.rows[0].$transit.save();
    await tbl.rows[1].$transit.save();
    await tbl.rows[2].$transit.save();
    await checkResult([row1, row2, row3]);

    // test primary key change
    row1.pk1 = 9;
    await tbl.rows[0].$transit.save();
    await checkResult([row1, row2, row3]);

    // test deleting a new row
    tbl.rows.push({});
    tbl.rows.pop();
    await sleep(1000);
    await checkResult([row1, row2, row3]);

    // test inserting an empty object
    // (tests identity management)
    let row4 : any = {};
    tbl.rows.push(row4);
    row4.pk1 = 10;
    row4.pk2 = 4;
    row4.str = "row 4";
    row4.num = 4;
    await row4.$transit.save();
    await checkResult([row1, row2, row3, row4]);

    expect(row4.$transit.isUnique()).toBe(true);
    row4.pk2 = 3;
    expect(row4.$transit.isUnique()).toBe(false);
    row4.pk2 = 4;

    // test reactivity
    let callCount = 0;
    window.app.$watch("simple.rows", function() { callCount++; }, {deep: true});
    await sleep(100);
    row1.$transit.saving = !row1.$transit.saving;
    await sleep(100);
    expect(callCount).toBe(1);
}, 10000);

test("Vue vueSave", async() => {
    var simple = window.app.simple;
    var ident = window.app.ident;

    await window.sqlserver.simple.truncate();
    await simple.execute();
    await window.sqlserver.ident.truncate();
    await ident.execute();

    while (!ident.complete) {
        await sleep(100);
    }

    simple.rows.push({ pk1: 10, pk2: 1, str: "row 1", num: null });
    let identRow = { str: "str" };
    ident.rows.push(identRow);

    await window.sqlserver.vueSave(simple, identRow);
    var simpleRows = await window.sqlserver.simple.select();
    var identRows = await window.sqlserver.ident.select();

    expect(simpleRows.length).toBe(1);
    expect(identRows.length).toBe(1);

    // make sure our new rows don't insert if saved again...
    await window.sqlserver.vueSave(simple);
});

test("Vue identity columns", async() => {
    var ident = window.app.ident;
    await window.sqlserver.ident.truncate();
    await ident.execute();

    ident.rows.push({ str: "str" });
    await ident.rows[0].$transit.save();

    let rows = await window.sqlserver.ident.select();
    expect(rows[0].str).toBe("str");

    // TODO: automatically updates row with updated values from server

    await ident.execute();

    ident.row.str = "str2";
    await ident.rows[0].$transit.save();

    rows = await window.sqlserver.ident.select();
    expect(rows[0].str).toBe("str2");
}, 10000);

test ("Vue extension APIs", async() => {
    window.app.callEcho.args = { msg: "foo" };
    await window.app.callEcho.execute();
    expect(window.app.callEcho.rows[0].message).toBe("foo");

    window.app.echoNum.args = { n: 5 };
    await window.app.echoNum.execute(5);
    expect(window.app.echoNum.result).toBe(5);

    await window.app.nothing.execute();
    expect(window.app.nothing.result).toBe(null);
});

test("Vue autosave", async() => {
    await window.sqlserver.simple.truncate();

    var tbl = window.app.simple;
    await tbl.execute();

    let row1 : any = { pk1: 10, pk2: 1, str: "row 1" };
    tbl.rows.push(row1);
    await row1.$transit.save();

    var saveCount = 0;
    row1.$transit.autosave(function(row) {
        saveCount++;
    });

    row1.str = "edit 1";
    await sleep(500);
    expect(saveCount).toBe(1);
    row1.str = "edit 2";
    await sleep(500);
    expect(saveCount).toBe(2);

    await row1.$transit.removeAndSave();
    row1.str = "edit 3";
    await sleep(500);
    expect(saveCount).toBe(2);

    await window.sqlserver.simple.truncate();
});

test("Vue CSV", async () => {
    var tbl = window.app.simple;
    await tbl.execute();
    while (tbl.rows.length) {
        await tbl.rows[0].$transit.removeAndSave();
    }

    tbl.rows.push({ pk1: 10, pk2: 1, str: "row 1", num: null });
    await tbl.rows[0].$transit.save();
    tbl.rows.push({ pk1: 10, pk2: 2, str: 'row,2\n"data"', num: 2 });
    await tbl.rows[0].$transit.save();

    let csv = window.app.simple.csv();
    expect(csv).toBe('pk1,pk2,str,num\n10,1,row 1,\n10,2,"row,2\n""data""",2\n');
});

function checkStatus(data, unused, inProgress, complete, error) {
    expect(data.unused).toBe(unused);
    expect(data.inProgress).toBe(inProgress);
    expect(data.complete).toBe(complete);
    expect(data.error == "").toBe(!error);
}

async function loadPage(url : string) : Promise<any> {
    let vConsole = new VirtualConsole();
    vConsole.sendTo(console);
    let error : Error | null = null;
    vConsole.on("jsdomError", (e) => { 
        error = e;
    });
    let dom = await JSDOM.fromURL(url, {
        runScripts: "dangerously",
        resources: "usable",
        virtualConsole: vConsole,
        includeNodeLocations: true,
    });
    return new Promise(function(resolve, reject) {
        dom.window.addEventListener("load", () => {
            if (error != null) {
                reject("Error during in-page JavaScript processing, see stack trace above");
            } else {
                resolve(dom.window);
            }
        });
    });
}

function sleep(ms) {
	return new Promise((resolve) => setTimeout(resolve, ms));
}
