import * as server from '../lib/Server';
import { JSDOM, VirtualConsole } from "jsdom";
import * as log from '../lib/Logger';

const PORT = 8003;
const configPath = "src/test/Vue/config2.json";

var serv;
var window;
var serv_sockets:any = [];
beforeEach(async function() {
    // inhibit the server-started log message...
    let log = console.log;
    console.log = () => {};
    serv = await server.startServer(configPath);
    // keep track of server connections so we can manually kill them later to shut down quickly
    serv.on("connection", (socket) => {
        serv_sockets.push(socket);
    });
    console.log = log;

    //inhibit the vue dev mode log message...
    let info = console.info;
    console.info = () => {};
    window = await loadPage(`http://localhost:${PORT}/vue2.html`);
    await sleep(1); // the vue log message triggers asynchronously after load, we need to wait one little bit longer....
    console.info = info;

    await window.initPromise;
});
afterEach(async function() {
    window.close();
    await new Promise((resolve, reject) => {
        serv.close(resolve);
        // close sockets manually for faster shutdown,
        // otherwise they wait until keep-alive timeout
        for (let s of serv_sockets) {
            s.destroy();
        }
        serv_sockets.length = 0;
    });
});

test("Vue loadNamespaceList", async () => {
    let list = await window.Transit.loadNamespaceList();
    expect(list).toEqual(["sqlserver"]);
});

test("Vue db proc", async () => {
    let success = 0;
    window.Transit.$sqlserver.$echo.onload = function() {
        success++;
    };

    checkStatus(window.data.echoStatus, true, false, false, false, false);
    window.data.echoArgs = {message: "test message"};
    await sleep(1);
    checkStatus(window.data.echoStatus, false, true, false, false, false);
    await sleep(100);
    expect(window.data.echo[0].message).toBe("test message");
    expect(success).toBe(1);
    checkStatus(window.data.echoStatus, false, false, false, true, false);
});

test("Vue function args", async () => {
    window.data.message = "functional message";
    await sleep(1000);
    expect(window.data.funcArgs[0].message).toBe("functional message");
});

test("Vue result reactivity", async () => {
    expect(window.app.oneRowArrayLength).toBe(0);
    await window.sqlserver.load("oneRowArray");
    expect(window.app.oneRowArrayLength).toBe(1);
});

test("Vue property naming", async () => {
    expect(window.data.foobar).not.toBeNull();
});


test("Vue db view", async () => {
    let resolved = false;
    checkStatus(window.data.All_DatatypesStatus, true, false, false, false, false);
    window.sqlserver.load("All_Datatypes").then(() => resolved = true);
    await sleep(1);
    checkStatus(window.data.All_DatatypesStatus, false, true, false, false, false);
    await sleep(100);
    expect(window.data.All_Datatypes["col bigint"]).toBe("3");
    expect(window.data.All_Datatypes["col varchar"]).toBe("xyz");
    expect(resolved).toBe(true);
    checkStatus(window.data.All_DatatypesStatus, false, false, false, true, false);

    window.data.All_DatatypesArgs.filter = { "col bigint": 5 } // will return 0 results
    await sleep(1000);
    expect(window.data.All_Datatypes["col bigint"]).toBe(undefined);
});


test("Vue RPC failure", async() => {
    // turn off error logging because we're gonna generate some
    (log as any).enabled = false;

    let error = 0;
    window.sqlserver.$echo.onloaderror = function() {
        error++;
    }

    // mute our unhandled init rejection...
    window.sqlserver.oninitPromise.catch(() => {});

    checkStatus(window.data.echoStatus, true, false, false, false, false);
    window.data.echoArgs.message = 1; // mismatched datatype which should trigger a server 500 error
    await sleep(1);
    checkStatus(window.data.echoStatus, false, true, false, false, false);
    await sleep(100);
    checkStatus(window.data.echoStatus, false, false, false, true, true);

    expect(error).toBe(1);

    (log as any).enabled = true;
});

test("Vue multiple calls", async() => {
    let origSend = window.XMLHttpRequest.prototype.send;
    let callCount = 0;
    let updateCount = 0;
    window.XMLHttpRequest.prototype.send = function() {
        callCount++;
        return origSend.apply(this, arguments);
    }
    window.sqlserver.$echo.onload = function() {
        updateCount++;
    }

    // trigger a request
    window.data.echoArgs.message = "test message 1";

    // trigger a request before the previous one has been sent
    window.data.echoArgs.message = "test message 2";
    await sleep(10);

    // trigger a request after the previous one has been sent
    window.data.echoArgs.message = "test message 3";

    await sleep(1000);
    expect(callCount).toBe(2);
    expect(updateCount).toBe(1);
    expect(window.data.echo[0].message).toBe("test message 3");
});

test("Vue table editing", async () => {
    let tbl = window.app.simple;
    let status = window.app.simpleStatus;
    let ns = window.sqlserver;

    async function checkResult(expected) {
        let actual = await window.sqlserver.api.simple.select(null, null, ["pk1","pk2"]);
        expect(actual.length).toBe(expected.length);
        for (var i = 0; i < actual.length; i++) {
            expect(expected[i]).toEqual(actual[i]);
        }
    }

    while (tbl.length) {
        tbl.splice(0,1);
        await window.sqlserver.save(tbl);
    }
    await checkResult([]);

    let row1 : any = { pk1: 10, pk2: 1, str: "row 1", num: null };
    let row2 : any = { pk1: 10, pk2: 2, str: "row 2", num: 1 };
    let row3 : any = { pk1: 10, pk2: 3, str: "row 3", num: 2 };

    expect(status.edited).toBe(false);
    await ns.load(tbl);
    expect(status.edited).toBe(false);

    // test insert, push
    tbl.push(row1);
    await sleep(1);
    expect(status.edited).toBe(true);
    await ns.save(tbl);
    await checkResult([row1]);
    expect(status.edited).toBe(false);


    // test update
    row1.str = "row 1 edited";
    await sleep(1);
    expect(status.edited).toBe(true);
    await ns.save(tbl);
    await checkResult([row1]);
    expect(status.edited).toBe(false);

    // test splice
    tbl.splice(0,0,row2,row3); // [1] -> [2, 3, 1]
    await sleep(1);
    expect(status.edited).toBe(true);
    await ns.save(tbl);
    await checkResult([row1, row2, row3]);
    expect(status.edited).toBe(false);
    expect(row1.$key).not.toBe(row2.$key);
    expect(row2.$key).not.toBe(row3.$key);
    expect(row3.$key).not.toBe(row1.$key);

    tbl.splice(1,1); // [2, 3, 1] -> [2, 1]
    await sleep(1);
    expect(status.edited).toBe(true);
    await ns.save(tbl);
    await checkResult([row1, row2]);
    expect(status.edited).toBe(false);

    // test pop
    tbl.pop(); // [2, 1] -> [2]
    await sleep(1);
    expect(status.edited).toBe(true);
    await ns.save(tbl);
    await checkResult([row2]);
    expect(status.edited).toBe(false);

    // test shift
    tbl.shift(); // [2] -> []
    await sleep(1);
    expect(status.edited).toBe(true);
    await ns.save(tbl);
    await checkResult([]);
    expect(status.edited).toBe(false);

    // test unshift
    tbl.unshift(row1, row2, row3); // [] -> [1, 2, 3]
    await sleep(1);
    expect(status.edited).toBe(true);
    await ns.save(tbl);
    await checkResult([row1, row2, row3]);
    expect(status.edited).toBe(false);

    // test primary key change
    row1.pk1 = 9;
    await sleep(1);
    expect(status.edited).toBe(true);
    await ns.save(tbl);
    await checkResult([row1, row2, row3]);
    expect(status.edited).toBe(false);

    // test deleting a new row
    tbl.push({});
    await sleep(1);
    expect(status.edited).toBe(true);
    tbl.pop();
    await sleep(1);
    expect(status.edited).toBe(true);
    await ns.save(tbl);
    await checkResult([row1, row2, row3]);
    expect(status.edited).toBe(false);

    // test inserting an empty object
    // (tests identity management)
    let row4 : any = {};
    tbl.push(row4); // [1, 2, 3] -> [1, 2, 3, 4]
    row4.pk1 = 10;
    row4.pk2 = 4;
    row4.str = "row 4";
    row4.num = 4;
    await sleep(1);
    expect(status.edited).toBe(true);
    await ns.save(tbl);
    await checkResult([row1, row2, row3, row4]);
    expect(status.edited).toBe(false);
    
    // test revert
    let oldRow4 = {};
    for (let n in row4) {
        oldRow4[n] = row4[n];
    }
    row4.str = "row 4 edited";
    tbl.splice(1,1);
    await sleep(1);
    expect(status.edited).toBe(true);
    ns.revert(tbl);
    await checkResult([row1, row2, row3, oldRow4]);
    expect(status.edited).toBe(false);

    // cancel recreates rows from scratch, so we need to update our row references.
    [row1, row2, row3, row4] = tbl;

    expect(ns.isUnique(row4)).toBe(true);
    row4.pk2 = 3;
    expect(ns.isUnique(row4)).toBe(false);
    row4.pk2 = 4;

    // clear out our data to set baseline for next test run.
    tbl.splice(0,tbl.length);
    ns.save(tbl);

}, 10000);

test("Vue table single row editing", async () => {
    let row = window.app.simpleSingle;
    let status = window.app.simpleSingleStatus;
    let ns = window.sqlserver;

    async function checkResult(expected) {
        let actual = await window.sqlserver.api.simple.select(null, null, ["pk1","pk2"]);
        expect(actual.length).toBe(expected.length);
        for (var i = 0; i < actual.length; i++) {
            expect(actual[i]).toEqual(expected[i]);
        }
    }

    await ns.truncate(row);
    await checkResult([]);
    
    let changeCount = 0;
    window.app.$watch("simpleSingle.str", function() {
        changeCount++;
    });

    expect(ns.count(row)).toBe(0);
    row.pk1 = 10;
    expect(ns.count(row)).toBe(1);
    row.pk2 = 1;
    row.str = "row 1";
    row.num = null;
    await ns.save(row);
    expect(ns.count(row)).toBe(1);
    expect(changeCount).toBe(1);
    await checkResult([row]);

    row.str = "edited";
    await ns.save(row);
    expect(ns.count(row)).toBe(1);
    expect(changeCount).toBe(2);
    await checkResult([row]);
}, 10000);

test("Vue truncate and save", async() => {
    let simple = window.app.simple;
    let ident = window.app.ident;
    let ns = window.sqlserver;

    await ns.truncate(simple, ident);

    simple.push({ pk1: 10, pk2: 1, str: "row 1", num: null });
    ident.push({ str: "str" });

    await ns.save(simple, ident);

    var simpleRows = await window.sqlserver.api.simple.select();
    var identRows = await window.sqlserver.api.ident.select();

    expect(simpleRows.length).toBe(1);
    expect(identRows.length).toBe(1);
});

test("Vue updating w/ identity columns", async() => {
    let ident = window.app.ident;
    let ns = window.sqlserver;

    await ns.api.ident.truncate();
    await ns.api.ident.insert(null, { str: "str" });
    await ns.load(ident);
    ident[0].str = "str edited";
    await ns.save(ident);

    var identRows = await window.sqlserver.api.ident.select();
    expect(identRows.length).toBe(1);
    expect(identRows[0].str).toBe("str edited");

    ident = window.app.identSingle;
    await ns.load(ident);
    ident.str = "str edited 2";
    await ns.save(ident); // make sure we don't try to update our identity column...

    var identRows = await window.sqlserver.api.ident.select();
    expect(identRows.length).toBe(1);
    expect(identRows[0].str).toBe("str edited 2");

});

test("Vue identity columns", async() => {
    var ident = window.app.ident;
    var ns = window.sqlserver;

    await ns.truncate(ident);

    ident.push({ str: "str" });
    await ns.save(ident);

    let rows = await ns.api.ident.select();
    expect(rows[0].str).toBe("str");

    // TODO: automatically updates row with updated values from server

    await ns.load(ident);

    ident[0].str = "str2";
    await ns.save(ident);

    rows = await ns.api.ident.select();
    expect(rows[0].str).toBe("str2");
}, 10000);

test("Vue large result size", async() => {

    // this test ALSO tests insertion into a table that only contains identity columns
    // and no other value columns.

	let ns = window.sqlserver;
	let many_rows = window.app.many_rows;
	let many_rows_edit = window.app.many_rows_edit;
    
    if (many_rows_edit[0].num != 0) {
        // cleanup from previous test
        many_rows_edit.unshift({num: 0});
        await ns.save(many_rows_edit);
        await ns.load(many_rows_edit);
    }

    let rows = await ns.api.many_rows.select();
	expect(many_rows.length).toBe(rows.length);
	expect(many_rows_edit.length).toBe(rows.length);
	expect(many_rows_edit[0].num).toBe(0);
	
	// make sure delete works as expected when editable
	many_rows_edit.splice(0,1);
	await ns.save(many_rows_edit);
	rows = await ns.api.many_rows.select(null, null, ["num"]);
	expect(rows.length).toBe(many_rows_edit.length);
	expect(rows[0].num).toBe(1);
	
	// reset back to original state
	many_rows_edit.unshift({num:0});
	await ns.save(many_rows_edit);
	rows = await ns.api.many_rows.select(null, null, ["num"]);
	expect(rows.length).toBe(many_rows_edit.length);
	expect(rows[0].num).toBe(0);
});

test("Vue result paging", async() => {
    let rows = window.app.many_rows_paged;
    let status = window.app.many_rows_pagedStatus;
    let options = window.app.many_rows_pagedArgs.options;

    expect(rows.length).toBe(1000);
    expect(rows[0].num).toBe(0);
    expect(rows[999].num).toBe(999);
    expect(status.pageCount).toBe(70);

    options.pageIndex = 31;
    await sleep(1000);
    expect(rows.length).toBe(1000);
    expect(rows[0].num).toBe(30000);
    expect(rows[999].num).toBe(30999);
    expect(status.pageCount).toBe(70);

    options.pageSize = 69999;
    options.pageIndex = 1;
    await sleep(1000);
    expect(rows.length).toBe(69999);
    expect(rows[0].num).toBe(0);
    expect(rows[69998].num).toBe(69998);
    expect(status.pageCount).toBe(2);

    options.pageIndex = 2;
    await sleep(1000);
    expect(rows.length).toBe(1);
    expect(rows[0].num).toBe(69999);
    expect(status.pageCount).toBe(2);
});

/* todo: allow extension APIs

test ("Vue extension APIs", async() => {
    window.app.callEchoArgs = { msg: "foo" };
    window.app.echoNumArgs = { n: 5 };

    await sleep(1000);

    expect(window.app.callEcho[0].message).toBe("foo");
    expect(window.app.echoNum.result).toBe(5);

    await window.app.nothing.execute();
    expect(window.app.nothing.result).toBe(undefined);
});

*/

test("Vue autosave", async() => {
    expect(window.autosaveCount).toBe(0);

    let ns = window.sqlserver;
    let tbl = window.app.autosave;
    await ns.truncate(tbl);
    expect(window.autosaveCount).toBe(1);

    await ns.load(tbl);
    let row1 : any = { pk1: 10, pk2: 1, str: "row 1" };
    tbl.push(row1);
    await sleep(500);
    expect(window.autosaveCount).toBe(2);
    
    row1.str = "edit 1";
    await sleep(500);
    expect(window.autosaveCount).toBe(3);
    let check = await ns.api.simple.select();
    expect(check[0].str).toBe("edit 1");

    row1.str = "edit 2";
    await sleep(500);
    expect(window.autosaveCount).toBe(4);
    check = await ns.api.simple.select();
    expect(check[0].str).toBe("edit 2");

    tbl.splice(0,1);
    await sleep(500);
    expect(window.autosaveCount).toBe(5);

    row1.str = "edit 3"; // row1 isn't in tbl anymore, shouldn't save
    await sleep(500);
    expect(window.autosaveCount).toBe(5);
    check = await ns.api.simple.select();
    expect(check.length).toBe(0);

    await ns.truncate(tbl);
});

test("Vue CSV", async () => {
    var tbl = window.app.simple;
    var ns = window.sqlserver;

    await ns.truncate(tbl);

    tbl.push({ pk1: 10, pk2: 1, str: "row 1", num: null });
    tbl.push({ pk1: 10, pk2: 2, str: 'row,2\n"data"', num: 2 });
    await ns.save(tbl);
    
    let csv = ns.csv(tbl);
    expect(csv).toBe('pk1,pk2,str,num\n10,1,row 1,\n10,2,"row,2\n""data""",2\n');
});

function checkStatus(data, unused, loading, saving, complete, error) {
    expect(data.unused).toBe(unused);
    expect(data.loading).toBe(loading);
    expect(data.saving).toBe(saving);
    expect(data.complete).toBe(complete);
    expect(data.error == "").toBe(!error);
}

async function loadPage(url : string) : Promise<any> {
    let vConsole = new VirtualConsole();
    vConsole.sendTo(console);
    let error : Error | null = null;
    vConsole.on("jsdomError", (e) => { 
        error = e;
    });
    let dom = await JSDOM.fromURL(url, {
        runScripts: "dangerously",
        resources: "usable",
        virtualConsole: vConsole,
        includeNodeLocations: true,
    });
    return new Promise(function(resolve, reject) {
        dom.window.addEventListener("load", () => {
            if (error != null) {
                reject("Error during in-page JavaScript processing, see stack trace above");
            } else {
                resolve(dom.window);
            }
        });
    });
}

function sleep(ms) {
	return new Promise((resolve) => setTimeout(resolve, ms));
}
